/*

  Copyright (C) 2000, 2001 Silicon Graphics, Inc.  All Rights Reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2.1 of the GNU Lesser General Public License 
  as published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement 
  or the like.  Any license provided herein, whether implied or 
  otherwise, applies only to this software file.  Patent licenses, if
  any, provided herein do not apply to combinations of this program with 
  other software, or any other product whatsoever.  

  You should have received a copy of the GNU Lesser General Public 
  License along with this program; if not, write the Free Software 
  Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, 
  USA.

  Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pky,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan

*/


#include "fmt.h"

static char _RCHK[][TYERROR] = {
/*UNK ADR I1 I2 I4 I8 R4 R8 CX DCX L1 L2 L4 L8 CH SB STR NML R16 QCX I16 */
 {0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0},/*UNUSED*/
 {0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*RET*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*REVERT*/ 
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*GOTO*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /* X */
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*SLSH*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*STCK*/


,{0,  0,  0, 0, 0, 0, 1, 1, 1,  1, 0, 0, 0, 0, 0, 0,  0,  0, 1,  1,  0} /* I */
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /* ED */
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*NED*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*IM*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*APOS*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*H*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*TL*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*TR*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*T*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*COLN*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*S*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*SP*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*SS*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*P*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*BN*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*BZ*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*F*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*E*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*EE*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*D*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*G*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*GE*/
,{0,  0,  0, 0, 0, 0, 1, 1, 1,  1, 0, 0, 0, 0, 1, 0,  0,  0, 1,  1,  1} /*L*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*A*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*AW*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*O*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*OM*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*Z*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*ZM*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*NONL*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*Q*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*EN*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*ENE*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*ES*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*ESE*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*B*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*BM*/
};


static char _WCHK[][TYERROR] = {
/*UNK ADR I1 I2 I4 I8 R4 R8 CX DCX L1 L2 L4 L8 CH SB STR NML R16 QCX I16 */
 {0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*UNUSED*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*RET*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*REVERT*/ 
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*GOTO*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /* X */
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*SLSH*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*STCK*/


,{0,  0,  0, 0, 0, 0, 1, 1, 1,  1, 0, 0, 0, 0, 0, 0,  0,  0, 1,  1,  0} /* I */
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /* ED */
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*NED*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*IM*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*APOS*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*H*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*TL*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*TR*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*T*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*COLN*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*S*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*SP*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*SS*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*P*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*BN*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*BZ*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*F*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*E*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*EE*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*D*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*G*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*GE*/
,{0,  0,  0, 0, 0, 0, 1, 1, 1,  1, 0, 0, 0, 0, 1, 0,  0,  0, 1,  1,  1} /*L*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*A*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*AW*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*O*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*OM*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*Z*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*ZM*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*NONL*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*Q*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*EN*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*ENE*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*ES*/
,{0,  0,  1, 1, 1, 1, 0, 0, 0,  0, 1, 1, 1, 1, 1, 0,  0,  0, 0,  0,  1} /*ESE*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*B*/
,{0,  0,  0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0,  0,  0, 0,  0,  0} /*BM*/
};
