/*
 * Copyright (C) 2012 Advanced Micro Devices, Inc.  All Rights Reserved.
 */

/*
 *  Copyright (C) 2006. QLogic Corporation. All Rights Reserved.
 */

/*
 * Copyright 2003, 2004, 2005, 2006 PathScale, Inc.  All Rights Reserved.
 */

/*

  Copyright (C) 2000, 2001 Silicon Graphics, Inc.  All Rights Reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement 
  or the like.  Any license provided herein, whether implied or 
  otherwise, applies only to this software file.  Patent licenses, if 
  any, provided herein do not apply to combinations of this program with 
  other software, or any other product whatsoever.  

  You should have received a copy of the GNU General Public License along
  with this program; if not, write the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston MA 02111-1307, USA.

  Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pky,
  Mountain View, CA 94043, or:

  http://www.sgi.com

  For further information regarding this notice, see:

  http://oss.sgi.com/projects/GenInfo/NoticeExplan

*/



/* USMID:  "\n@(#)5.0_pl/headers/intrin.h	5.24	10/26/99 15:47:10\n" */

# define I1_MASK               (1<<Integer_1)
# define I2_MASK               (1<<Integer_2)
# define I4_MASK               (1<<Integer_4)
# define I8_MASK               (1<<Integer_8)
# define S4_MASK               (1<<Real_4)
# define S8_MASK               (1<<Real_8)
# define S16_MASK              (1<<Real_16)
# define C4_MASK               (1<<Complex_4)
# define C8_MASK               (1<<Complex_8)
# define C16_MASK              (1<<Complex_16)
# define L1_MASK               (1<<Logical_1)
# define L2_MASK               (1<<Logical_2)
# define L4_MASK               (1<<Logical_4)
# define L8_MASK               (1<<Logical_8)
# define P_MASK                (1<<CRI_Ptr_8)
# ifdef _EXTENDED_CRI_CHAR_POINTER
# define CP_MASK               (1<<CRI_Ch_Ptr_8)
# else
# define CP_MASK               (0)
# endif
# define X_MASK                (1<<Structure_Type)
# define c_MASK                (1<<Short_Char_Const)

# define TYPELESS_MASK         (1<<Typeless_1|1<<Typeless_2|                  \
                                1<<Typeless_4|1<<Typeless_8|                  \
                                1<<Short_Typeless_Const)

# define REAL_MASK             (S4_MASK|S8_MASK|S16_MASK)

# define COMPLEX_MASK          (C4_MASK|C8_MASK|C16_MASK)

# define INTEGER_MASK          (I1_MASK|I2_MASK|I4_MASK|I8_MASK)

# define LOGICAL_MASK          (L1_MASK|L2_MASK|L4_MASK|L8_MASK)

# define CHARACTER_MASK        (1<<Character_1|                               \
                                1<<Character_2|                               \
                                1<<Character_4|                               \
                                c_MASK)

# define INTEGER_COMPLEX_MASK  (INTEGER_MASK|COMPLEX_MASK)
# define INTEGER_REAL_MASK     (INTEGER_MASK|REAL_MASK)
# define INTEGER_TYPELESS_MASK (INTEGER_MASK|TYPELESS_MASK)
# define INTEGER_POINTER_MASK  (INTEGER_MASK|P_MASK)
# define REAL_COMPLEX_MASK     (REAL_MASK|COMPLEX_MASK)
# define IRC_MASK              (INTEGER_REAL_MASK|COMPLEX_MASK)
# define IRT_MASK              (INTEGER_REAL_MASK|TYPELESS_MASK)
# define IRP_MASK              (IRT_MASK|LOGICAL_MASK|P_MASK|c_MASK)
# define IRCL_MASK             (INTEGER_REAL_MASK|COMPLEX_MASK|LOGICAL_MASK)

# define IRCLC_MASK            (INTEGER_REAL_MASK|                            \
                                COMPLEX_MASK|                                 \
                                LOGICAL_MASK|                                 \
                                CHARACTER_MASK)


# define ALL_MASK              (INTEGER_MASK|                                 \
                                REAL_MASK|                                    \
                                COMPLEX_MASK|                                 \
                                LOGICAL_MASK|                                 \
                                CHARACTER_MASK|                               \
                                X_MASK|                                       \
                                P_MASK)

/* Map name of intrinsic family onto mask used within "families" member of
 * intrin_tbl_entry. Must be in order by name. */
intrin_family_t intrin_families[] = {
  { "ANSI", ANSI_FAMILY },
  { "EVERY", EVERY_FAMILY },
  { "G77", G77_FAMILY },
  { "OMP", OMP_FAMILY },
  { "PGI", PGI_FAMILY },
  { "TRADITIONAL", TRADITIONAL_FAMILY }
  };
int SIZEOF_INTRIN_FAMILIES = (sizeof intrin_families)/(sizeof *intrin_families);

/**************************************************************************\
|* The global Intrinsic table                                             *|
\**************************************************************************/
#ifdef KEY
/*
 * How to decipher this table:
 * (I inferred this from reading code, so it might not be authoritative.)
 *
 * 1. The unindented entries must be upper case, alphabetized by the (initial)
 *    field id_str
 *
 * 2. To implement a function and an intrinsic having the same name, make
 *    an ordinary entry for the function and an additional entry for the
 *    subroutine with INTRIN_SUBR_SUFFIX appended to the name. Be careful
 *    about alphabetization.
 *
 * 3. The unindented entry provides the name by which the user calls the
 *    intrinsic; the singly-indented entry provides information about a
 *    particular implementation; and the doubly-indented entries provide
 *    information about each of the arguments for that implementation.
 *    One might assume that the unindented entry is a generic and the
 *    singly-indented entries are specifics using the ANSI meanings of
 *    those terms, but not so. Generics and specifics alike start
 *    with an unindented entry; a generic has "passable" set to false and
 *    accepts a variety of data types (but then so do some intrinsics which
 *    aren't generics.)
 *
 * 4. In the unindented entry:
 *    id_str: name the user types (but see #1 above)
 *    name_len: length of id_str (didn't they know about "strlen"?)
 *    elemental:
 *    function: 1 if intrinsic is function
 *    passable: allowed to pass these to dummy args of type "external"
 *    external:
 *    optional:
 *    dope:
 *    generic: always 1, even for a specific like ALOG10--doesn't mean ANSI
 *      generic, means "unindented entry"!
 *    enabled: initially 0. At execution time, after scanning the command-line
 *      options, we set this true for every intrinsic which the compiler
 *      should recognize
 *    n_specifics: number of singly-indented entries under this one
 *    intrin_enum: 0
 *    data_type: 0
 *    families: Indicates which "families" this intrinsic belongs to:
 *      ANSI_FAMILY | PGI_FAMILY | G77_FAMILY, etc. At execution time, we "or"
 *      EVERY_FAMILY with the value which appears in the table source code
 *
 * 5. In the singly-indented entry:
 *    id_str: if "external" is set, name of the library function
 *      else same as unindented entry
 *    name_len: length of id_str (didn't they know about "strlen"?)
 *    elemental:
 *    function: 1 if intrinsic is function
 *    passable:
 *    external: 0 if ever implemented inline, 1 if always uses library function
 *    optional:
 *    dope: 1 if result is an array
 *    generic: 0
 *    enabled: ignored
 *    n_specifics:
 *    intrin_enum: Member of intrinsic_values enum which this implementation
 *     uses to dispatch, via table intrinsic_semantics, to a function which
 *     performs semantics work for this intrinsic
 *    data_type: Result type, using constants like Real_4. Often this is 0;
 *     not clear whether that represents an error. It is an error in the
 *     case where this entry represents a function in the "Specific
 *     names for intrinsic functions" table in the standard (that is, a
 *     function which can be passed as an actual argument): without an
 *     explicit type, the front end gives spurious warnings and reads
 *     variables containing trash. Otherwise, 0 magically works (so far.)
 *    families: ignored
 *
 * 6. In the doubly-indented entry:
 *    id_str: dummy argument name
 *    name_len: length of id_str (didn't they know about "strlen"?)
 *    elemental:
 *    function: 1 if this argument is a dummy function or subroutine
 *    passable:
 *    external:
 *    optional:1 if this argument is optional
 *    dope: if n_specifics is nonzero, this is 1 to pass dope vector; else 0
 *    generic: 0
 *    enabled: ignored
 *    n_specifics: 0 for scalar; else rank of array
 *    intrin_enum: 0
 *    data_type: Mask of acceptable data types for this arg, e.g. S4_MASK;
 *      or 0 if this argument is an "external"
 *    families: ignored
 *
 * For intrinsics implemented via external procedures, arguments are generally
 * passed by reference. A Fortran intrinsic function which returns character
 * corresponds to a C "void" function whose first two arguments are a "char *"
 * and an "int" (containing the length of the destination string.) When an
 * intrinsic procedure takes a character argument, the corresponding C
 * function takes a "char *" plus an additional "int" (containing the length
 * of the string) but the "int" is placed at the end of the argument list,
 * after any other arguments.
 *
 * When an argument is optional and the intrinsic is implemented via an
 * external procedure, the generated code passes a nil pointer when the
 * argument is missing.
 *
 * User-written Fortran procedures generate linker symbols containing lowercase
 * letters ending in an underscore (two underscores if the Fortran name
 * contains an underscore) so C functions which implement intrinsics should
 * generally use some naming scheme that can't conflict with that one: e.g.
 * uppercase letters or names not ending in underscore.
 */
#endif /* KEY */

intrin_tbl_type         intrin_tbl[] =

{

{"PLACEHOLDER-NOTUSED",0,0,0,0,0,0,0,0,0,0,0,0,0},
#ifdef KEY /* Bug 1683 */
{"ABORT",5,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"pathf90_abort",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
#endif /* KEY Bug 1683 */
{"ABS",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ABS",3,1,1,0,0,0,0,0,0,0,Abs_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
# ifdef _NEED_AT_SIGN_INTRINSICS
{"ABS@",4,1,1,1,0,0,0,1,0,1,0,0,0},
   {"ABS@",4,1,1,0,0,0,0,0,0,0,Abs_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
# endif
#ifdef KEY /* Bug 1683 */
{"ACCESS",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"pathf90_access",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
     {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
     {"MODE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"ACHAR",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ACHAR",5,1,1,0,0,0,0,0,0,0,Achar_Intrinsic,Character_1,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ACOS",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ACOS",4,1,1,0,0,0,0,0,0,0,Acos_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ACOSD",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ACOSD",5,1,1,0,0,0,0,0,1,0,Acosd_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ADD_AND_FETCH",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"ADD_AND_FETCH",13,1,1,0,0,0,0,0,1,0,Add_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"ADD_AND_FETCH",13,1,1,0,0,0,0,0,1,0,Add_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"ADJUSTL",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ADJUSTL",7,1,1,0,0,0,0,0,0,0,Adjustl_Intrinsic,0,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"ADJUSTR",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ADJUSTR",7,1,1,0,0,0,0,0,0,0,Adjustr_Intrinsic,0,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"AIMAG",5,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 9060 */
   {"AIMAG",5,1,1,0,0,0,0,0,0,0,Aimag_Intrinsic,Real_4,0},
#else /* KEY Bug 9060 */
   {"AIMAG",5,1,1,0,0,0,0,0,0,0,Aimag_Intrinsic,0,0},
#endif /* KEY Bug 9060 */
      {"Z",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
{"AINT",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"AINT",4,1,1,0,0,0,0,0,0,0,Aint_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"ALARM",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_alarm",13,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"SECONDS",7,0,0,0,0,0,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"HANDLER",7,0,1,0,0,0,0,0,0,0,0,0,0},
      {"STATUS",7,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"ALARM" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_alarm",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"SECONDS",7,0,0,0,0,0,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"HANDLER",7,0,1,0,0,0,0,0,0,0,0,0,0},
      {"STATUS",7,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"ALL",3,0,1,0,0,0,0,1,0,56,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,1,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,2,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,3,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,4,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,5,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,6,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,1,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,2,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,3,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,4,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,5,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,6,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,1,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,2,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,3,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,4,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,5,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,6,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L1_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L1_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L1_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L1_MASK,0},
   {"_ALL_1",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L1_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L2_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L2_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L2_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L2_MASK,0},
   {"_ALL_2",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L2_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L4_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L4_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L4_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L4_MASK,0},
   {"_ALL_4",6,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L4_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,1,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,2,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,3,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,4,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,5,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,6,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L8_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L8_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L8_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L8_MASK,0},
   {"_ALL",4,0,1,0,1,0,1,0,0,0,All_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L8_MASK,0},
{"ALLOCATED",9,0,1,0,1,0,0,1,0,7,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ALLOCATED",10,0,1,0,1,0,0,0,0,0,Allocated_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
{"ALOG",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ALOG",4,1,1,0,0,0,0,0,0,0,Alog_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ALOG10",6,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ALOG10",6,1,1,0,0,0,0,0,0,0,Alog10_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"AMAX0",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"AMAX0",5,1,1,0,0,0,0,0,0,0,Amax0_Intrinsic,Integer_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"AMAX1",5,1,1,0,0,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"AMAX1",5,1,1,0,0,0,0,0,0,0,Amax1_Intrinsic,Real_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
   {"AMAX1",5,1,1,0,0,0,0,0,0,0,Amax1_Intrinsic,Real_8,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
{"AMIN0",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"AMIN0",5,1,1,0,0,0,0,0,0,0,Amin0_Intrinsic,Integer_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"AMIN1",5,1,1,0,0,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"AMIN1",5,1,1,0,0,0,0,0,0,0,Amin1_Intrinsic,Real_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
   {"AMIN1",5,1,1,0,0,0,0,0,0,0,Amin1_Intrinsic,Real_8,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
{"AMOD",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"AMOD",4,1,1,0,0,0,0,0,0,0,Amod_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"AND",3,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 1683 */
   /* For g77 compatibility, must treat values bitwise regardless of type */
   {"AND",3,1,1,0,0,0,0,0,1,0,Iand_Intrinsic,0,0},
#else
   {"AND",3,1,1,0,0,0,0,0,1,0,And_Intrinsic,0,0},
#endif /* KEY Bug 1683 */
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"AND_AND_FETCH",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"AND_AND_FETCH",13,1,1,0,0,0,0,0,1,0,And_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"AND_AND_FETCH",13,1,1,0,0,0,0,0,1,0,And_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"ANINT",5,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 3410 */
   {"ANINT",5,1,1,0,0,0,0,0,0,0,Anint_Intrinsic,Real_4,0},
#else
   {"ANINT",5,1,1,0,0,0,0,0,0,0,Anint_Intrinsic,0,0},
#endif /* KEY Bug 3410 */
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"ANY",3,0,1,0,0,0,0,1,0,56,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,1,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,2,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,3,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,4,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,5,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,6,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,1,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,2,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,3,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,4,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,5,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,6,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,1,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,2,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,3,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,4,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,5,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,6,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L1_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L1_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L1_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L1_MASK,0},
   {"_ANY_1",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L1_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L2_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L2_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L2_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L2_MASK,0},
   {"_ANY_2",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L2_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L4_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L4_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L4_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L4_MASK,0},
   {"_ANY_4",6,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L4_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,1,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,2,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,3,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,4,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,5,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,6,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L8_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L8_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L8_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L8_MASK,0},
   {"_ANY",4,0,1,0,1,0,1,0,0,0,Any_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L8_MASK,0},
{"ASIN",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ASIN",4,1,1,0,0,0,0,0,0,0,Asin_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ASIND",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ASIND",5,1,1,0,0,0,0,0,1,0,Asind_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ASSOCIATED",10,0,1,0,1,0,0,1,0,144,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_4",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_ASSOCIATED_8",13,0,1,0,1,0,0,0,0,0,Associated_Intrinsic,0,0},
      {"POINTER",7,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"TARGET",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
{"ATAN",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ATAN",4,1,1,0,0,0,0,0,0,0,Atan_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ATAN2",5,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ATAN2",5,1,1,0,0,0,0,0,0,0,Atan2_Intrinsic,Real_4,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ATAN2D",6,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ATAN2D",6,1,1,0,0,0,0,0,1,0,Atan2d_Intrinsic,Real_4,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ATAND",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ATAND",5,1,1,0,0,0,0,0,1,0,Atand_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
#ifdef KEY /* Bug 1683 */
{"BESJ0",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_besj0_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
  {"G77_dbesj0_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"BESJ1",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_besj1_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
  {"G77_dbesj1_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"BESJN",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_besjn_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
  {"G77_dbesjn_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"BESY0",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_besy0_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
  {"G77_dbesy0_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"BESY1",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_besy1_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
  {"G77_dbesy1_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"BESYN",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_besyn_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
  {"G77_dbesyn_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
#endif /* KEY Bug 1683 */
{"BITEST",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"BITEST",6,1,1,0,0,0,0,0,0,0,Bitest_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"BIT_SIZE",8,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"BIT_SIZE",8,1,1,0,0,0,0,0,0,0,Bit_Size_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"BJTEST",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"BJTEST",6,1,1,0,0,0,0,0,0,0,Bjtest_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"BKTEST",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"BKTEST",6,1,1,0,0,0,0,0,0,0,Bktest_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"BTEST",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"BTEST",5,1,1,0,0,0,0,0,0,0,Btest_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"CABS",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CABS",4,1,1,0,0,0,0,0,0,0,Cabs_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
{"CCOS",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CCOS",4,1,1,0,0,0,0,0,0,0,Ccos_Intrinsic,Complex_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
#ifdef KEY
{"CDABS",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#else
{"CDABS",5,1,1,0,0,0,0,1,0,1,0,0,0},
#endif
   {"CDABS",5,1,1,0,0,0,0,0,1,0,Cdabs_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
#ifdef KEY
{"CDCOS",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#else
{"CDCOS",5,1,1,0,0,0,0,1,0,1,0,0,0},
#endif
   {"CDCOS",5,1,1,0,0,0,0,0,1,0,Cdcos_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
#ifdef KEY
{"CDEXP",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#else
{"CDEXP",5,1,1,0,0,0,0,1,0,1,0,0,0},
#endif
   {"CDEXP",5,1,1,0,0,0,0,0,1,0,Cdexp_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
#ifdef KEY
{"CDLOG",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#else
{"CDLOG",5,1,1,0,0,0,0,1,0,1,0,0,0},
#endif
   {"CDLOG",5,1,1,0,0,0,0,0,1,0,Cdlog_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
#ifdef KEY
{"CDSIN",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#else
{"CDSIN",5,1,1,0,0,0,0,1,0,1,0,0,0},
#endif
   {"CDSIN",5,1,1,0,0,0,0,0,1,0,Cdsin_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
#ifdef KEY
{"CDSQRT",6,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#else
{"CDSQRT",6,1,1,0,0,0,0,1,0,1,0,0,0},
#endif
   {"CDSQRT",6,1,1,0,0,0,0,0,1,0,Cdsqrt_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"CEILING",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"CEILING",7,1,1,0,0,0,0,0,0,0,Ceiling_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"CEXP",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CEXP",4,1,1,0,0,0,0,0,0,0,Cexp_Intrinsic,Complex_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
{"CHAR",4,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CHAR",4,1,1,0,0,0,0,0,0,0,Char_Intrinsic,Character_1,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"CHDIR",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"pathf90_chdir",13,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"DIR",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"CHDIR" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
  {"pathf90_chdir",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"DIR",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"CHMOD",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"pathf90_chmod",13,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"MODE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"CHMOD" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
  {"pathf90_chmod",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"MODE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"CLEAR_IEEE_EXCEPTION",20,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CLEAR_IEEE_EXCEPTION",20,1,0,0,0,0,0,0,0,0,
                                     Clear_Ieee_Exception_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"EXCEPTION",9,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"EXCEPTION",9,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"CLOC",4,0,1,0,0,0,0,1,0,8,0,0,TRADITIONAL_FAMILY},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,0,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,1,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,2,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,3,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,4,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,5,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,6,0,CHARACTER_MASK,0},
   {"CLOC",4,0,1,0,0,0,0,0,1,0,Cloc_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,1,0,0,7,0,CHARACTER_MASK,0},
{"CLOCK",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_F90_CLOCK_",11,0,1,0,1,0,0,0,1,0,Clock_Intrinsic,0,0},
# else
   {"_CLOCK",6,0,1,0,1,0,0,0,1,0,Clock_Intrinsic,TYPELESS_DEFAULT_TYPE,0},
# endif
{"CLOG",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CLOG",4,1,1,0,0,0,0,0,0,0,Clog_Intrinsic,Complex_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
{"CMPLX",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CMPLX",5,1,1,0,0,0,0,0,0,0,Cmplx_Intrinsic,Complex_4,0},
#ifdef KEY /* Bug 12482 */
      {"X",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK|TYPELESS_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,IRC_MASK|TYPELESS_MASK,0},
#else /* KEY Bug 12482 */
      {"X",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,IRC_MASK,0},
#endif /* KEY Bug 12482 */
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"COMMAND_ARGUMENT_COUNT",22,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY},
   {"_Command_argument_count",23,0,1,0,1,0,0,0,0,0,Pathf90_Intrinsic,Integer_4,0},
{"COMPARE_AND_SWAP",16,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"COMPARE_AND_SWAP",16,1,1,0,0,0,0,0,1,0,Compare_And_Swap_Intrinsic,
                                                                 Logical_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"COMPARE_AND_SWAP",16,1,1,0,0,0,0,0,1,0,Compare_And_Swap_Intrinsic,
                                                                 Logical_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"COMPL",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"COMPL",5,1,1,0,0,0,0,0,1,0,Compl_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"CONJG",5,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 9060 */
   {"CONJG",5,1,1,0,0,0,0,0,0,0,Conjg_Intrinsic,Complex_4,0},
#else /* KEY Bug 9060 */
   {"CONJG",5,1,1,0,0,0,0,0,0,0,Conjg_Intrinsic,0,0},
#endif /* KEY Bug 9060 */
      {"Z",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
{"COS",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"COS",3,1,1,0,0,0,0,0,0,0,Cos_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_COMPLEX_MASK,0},
{"COSD",4,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"COSD",4,1,1,0,0,0,0,0,1,0,Cosd_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"COSH",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"COSH",4,1,1,0,0,0,0,0,0,0,Cosh_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"COT",3,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"COT",3,1,1,0,0,0,0,0,0,0,Cot_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"COTAN",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"COTAN",5,1,1,0,0,0,0,0,0,0,Cot_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"COUNT",5,0,1,0,0,0,0,1,0,56,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,1,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,2,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,3,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,4,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,5,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,6,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L1_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,1,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,2,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,3,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,4,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,5,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,6,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L2_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,1,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,2,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,3,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,4,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,5,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,6,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L4_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L1_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L1_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L1_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L1_MASK,0},
   {"_COUNT_1",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L1_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L2_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L2_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L2_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L2_MASK,0},
   {"_COUNT_2",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L2_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L4_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L4_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L4_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L4_MASK,0},
   {"_COUNT_4",8,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L4_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,1,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,2,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,3,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,4,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,5,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,6,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L8_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,L8_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,L8_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,L8_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,L8_MASK,0},
   {"_COUNT",6,0,1,0,1,0,1,0,0,0,Count_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,L8_MASK,0},
{"CPU_TIME",8,0,0,0,1,0,0,1,0,3,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"_CPU_TIME_4",11,0,0,0,1,0,0,0,0,0,0,0,0},
      {"TIME",4,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
   {"_CPU_TIME_8",11,0,0,0,1,0,0,0,0,0,0,0,0},
      {"TIME",4,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
   {"_CPU_TIME_16",12,0,0,0,1,0,0,0,0,0,0,0,0},
      {"TIME",4,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"CQABS",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CQABS",5,1,1,0,0,0,0,0,1,0,Cqabs_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"CQCOS",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CQCOS",5,1,1,0,0,0,0,0,1,0,Cqcos_Intrinsic,Complex_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"CQEXP",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CQEXP",5,1,1,0,0,0,0,0,1,0,Cqexp_Intrinsic,Complex_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"CQLOG",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CQLOG",5,1,1,0,0,0,0,0,1,0,Cqlog_Intrinsic,Complex_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"CQSIN",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CQSIN",5,1,1,0,0,0,0,0,1,0,Cqsin_Intrinsic,Complex_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"CQSQRT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CQSQRT",6,1,1,0,0,0,0,0,1,0,Cqsqrt_Intrinsic,Complex_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"CSHIFT",6,0,1,0,0,0,0,1,0,13,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,1,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,2,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,3,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,4,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,5,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,6,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,7,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,2,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,3,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,2,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,4,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,3,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,5,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,4,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,6,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,5,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_CSHIFT",7,0,1,0,1,0,1,0,0,7,Cshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,6,0,INTEGER_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"CSIN",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CSIN",4,1,1,0,0,0,0,0,0,0,Csin_Intrinsic,Complex_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
{"CSMG",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CSMG",4,1,1,0,0,0,0,0,1,0,Csmg_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"CSQRT",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"CSQRT",5,1,1,0,0,0,0,0,0,0,Csqrt_Intrinsic,Complex_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
#ifdef KEY /* Bug 1683 */
{"CTIME",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_ctime4",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Character_1,0},
      {"STIME",5,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"pathf90_ctime8",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Character_1,0},
      {"STIME",5,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"CTIME" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,2,0,0,G77_FAMILY},
   {"pathf90_subr_ctime4",19,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"STIME",5,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"RESULT",6,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
   {"pathf90_subr_ctime8",19,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"STIME",5,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"RESULT",6,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"CVMGM",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CVMGM",5,1,1,0,0,0,0,0,1,0,Cvmgm_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"CVMGN",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CVMGN",5,1,1,0,0,0,0,0,1,0,Cvmgn_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"CVMGP",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CVMGP",5,1,1,0,0,0,0,0,1,0,Cvmgp_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"CVMGT",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CVMGT",5,1,1,0,0,0,0,0,1,0,Cvmgt_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,LOGICAL_MASK,0},
{"CVMGZ",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"CVMGZ",5,1,1,0,0,0,0,0,1,0,Cvmgz_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"C_LOC",5,0,1,0,0,0,0,1,0,9,0,0,TRADITIONAL_FAMILY},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,1,0,0,0,0,0,0,0,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,0,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,1,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,2,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,3,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,4,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,5,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,6,0,(ALL_MASK|CP_MASK),0},
   {"C_LOC",5,0,1,0,0,0,0,0,1,0,C_Loc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,1,0,0,7,0,(ALL_MASK|CP_MASK),0},
{"DABS",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DABS",4,1,1,0,0,0,0,0,0,0,Dabs_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DACOS",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DACOS",5,1,1,0,0,0,0,0,0,0,Dacos_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DACOSD",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DACOSD",6,1,1,0,0,0,0,0,0,0,Dacosd_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DASIN",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DASIN",5,1,1,0,0,0,0,0,0,0,Dasin_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DASIND",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DASIND",6,1,1,0,0,0,0,0,0,0,Dasind_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DATAN",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DATAN",5,1,1,0,0,0,0,0,0,0,Datan_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DATAN2",6,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DATAN2",6,1,1,0,0,0,0,0,0,0,Datan2_Intrinsic,Real_8,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DATAN2D",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DATAN2D",7,1,1,0,0,0,0,0,0,0,Datan2d_Intrinsic,Real_8,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DATAND",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DATAND",6,1,1,0,0,0,0,0,0,0,Datand_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DATE",4,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_F90_DATE_",10,0,1,0,1,0,0,0,1,0,Date_Intrinsic,0,0},
# else
   {"_DATE",5,0,1,0,1,0,0,0,1,0,Date_Intrinsic,TYPELESS_DEFAULT_TYPE,0},
# endif
#ifdef KEY /* Bug 1683 */
/* g77 version is a subroutine and behaves slightly differently */
{"DATE" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_date",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"DATE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"DATE_AND_TIME",13,0,0,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_F90_DATE_AND_TIME",18,0,0,0,1,0,0,0,0,0,0,0,0},
# else
   {"_DATE_AND_TIME",14,0,0,0,1,0,0,0,0,0,0,0,0},
# endif
      {"DATE",4,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
      {"TIME",4,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
      {"ZONE",4,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
      {"VALUES",6,0,0,0,0,1,1,0,0,1,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"DBESJ0",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_dbesj0_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DBESJ1",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_dbesj1_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DBESJN",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_dbesjn_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DBESY0",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_dbesy0_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DBESY1",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_dbesy1_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DBESYN",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
  {"G77_dbesyn_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
#endif /* KEY Bug 1683 */
{"DBLE",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DBLE",4,1,1,0,0,0,0,0,0,0,Dble_Intrinsic,Real_8,0},
#ifdef KEY /* Bug 12482 */
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK|TYPELESS_MASK,0},
#else /* KEY Bug 12482 */
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
#endif /* KEY Bug 12482 */
{"DBLEQ",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DBLEQ",5,1,1,0,0,0,0,0,0,0,Dbleq_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"DCMPLX",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DCMPLX",6,1,1,0,0,0,0,0,1,0,Dcmplx_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,IRC_MASK,0},
{"DCONJG",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DCONJG",6,1,1,0,0,0,0,0,1,0,Dconjg_Intrinsic,Complex_8,0},
      {"Z",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"DCOS",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DCOS",4,1,1,0,0,0,0,0,0,0,Dcos_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DCOSD",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DCOSD",5,1,1,0,0,0,0,0,0,0,Dcosd_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DCOSH",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DCOSH",5,1,1,0,0,0,0,0,0,0,Dcosh_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DCOT",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DCOT",4,1,1,0,0,0,0,0,0,0,Dcot_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DCOTAN",6,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DCOTAN",6,1,1,0,0,0,0,0,0,0,Dcot_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DDIM",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DDIM",4,1,1,0,0,0,0,0,0,0,Ddim_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
#ifdef KEY /* Bug 1324 */
{"DERF",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DERF",4,1,1,0,0,0,0,0,1,0,Derf_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"DERFC",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DERFC",5,1,1,0,0,0,0,0,1,0,Derfc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
#endif /* KEY Bug 1324 */
{"DEXP",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DEXP",4,1,1,0,0,0,0,0,0,0,Dexp_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DFLOAT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DFLOAT",6,1,1,0,0,0,0,0,1,0,Dfloat_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"DFLOATI",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DFLOATI",7,1,1,0,0,0,0,0,1,0,Dfloati_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"DFLOATJ",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DFLOATJ",7,1,1,0,0,0,0,0,1,0,Dfloatj_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"DFLOATK",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DFLOATK",7,1,1,0,0,0,0,0,1,0,Dfloatk_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"DIGITS",6,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DIGITS",6,1,1,0,0,0,0,0,0,0,Digits_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
{"DIM",3,1,1,1,0,0,1,1,0,4,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 9060 */
   /* Code in p_utils.c function parse_deref assumes that when the unindented
    * line represents a generic intrinsic interface, the first entry is the
    * specific intrinsica having the same name. Not clear whether that's
    * intentional or an oversight, but alphabetizing these indented entries
    * satisfies that assumption in this particular case. :-) */
#else
   {"IDIM",4,1,1,0,0,0,0,0,0,0,Dim_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
#endif /* KEY Bug 9060 */
# ifdef _TARGET_OS_MAX
   {"_HDIM_",6,1,1,0,1,0,0,0,0,0,Dim_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
   {"_DIM_",5,1,1,0,1,0,0,0,0,0,Dim_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
# else
   {"DIM",3,1,1,0,0,0,0,0,0,0,Dim_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
   {"DDIM",4,1,1,0,0,0,0,0,0,0,Dim_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
# endif
#ifdef KEY /* Bug 9060 */
   {"IDIM",4,1,1,0,0,0,0,0,0,0,Dim_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
#endif /* KEY Bug 9060 */
   {"QDIM",4,1,1,0,0,0,0,0,0,0,Dim_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"DIMAG",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DIMAG",5,1,1,0,0,0,0,0,0,0,Dimag_Intrinsic,Real_8,0},
      {"Z",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"DINT",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DINT",4,1,1,0,0,0,0,0,0,0,Dint_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DISABLE_IEEE_INTERRUPT",22,1,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DISABLE_IEEE_INTERRUPT",22,1,0,0,0,0,0,0,0,0,
                                     Disable_Ieee_Interrupt_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"INTERRUPT",9,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"INTERRUPT",9,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"DLOG",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DLOG",4,1,1,0,0,0,0,0,0,0,Dlog_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DLOG10",6,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DLOG10",6,1,1,0,0,0,0,0,0,0,Dlog10_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DMAX1",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DMAX1",5,1,1,0,0,0,0,0,0,0,Dmax1_Intrinsic,Real_8,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
{"DMIN1",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DMIN1",5,1,1,0,0,0,0,0,0,0,Dmin1_Intrinsic,Real_8,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
{"DMOD",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DMOD",4,1,1,0,0,0,0,0,0,0,Dmod_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DNINT",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DNINT",5,1,1,0,0,0,0,0,0,0,Dnint_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DOT_PRODUCT",11,0,1,0,0,0,0,1,0,116,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_DOT_C4C4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_C4C",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_C4Z",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_C4S4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_C4S",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_C4D",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_C4I1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_C4I2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_C4I4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_C4J",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_CC4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_CC",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_CZ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_CS4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_CS",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_CD",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_CI1",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_CI2",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_CI4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_CJ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_ZC4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_ZC",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_ZZ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_ZS4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_ZS",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_ZD",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_ZI1",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_ZI2",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_ZI4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_ZJ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_S4C4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_S4C",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_S4Z",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_S4S4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_S4S",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_S4D",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_S4I1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_S4I2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_S4I4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_S4J",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_SC4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_SC",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_SZ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_SS4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_SS",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_SD",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_SI1",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_SI2",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_SI4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_SJ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_DC4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_DC",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_DZ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_DS4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_DS",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_DD",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_DI1",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_DI2",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_DI4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_DJ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_I1C4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_I1C",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_I1Z",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_I1S4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_I1S",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_I1D",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_I1I1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_1,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_I1I2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_2,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_I1I4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_I1J",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_I2C4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_I2C",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_I2Z",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_I2S4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_I2S",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_I2D",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_I2I1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_2,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_I2I2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_2,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_I2I4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_I2J",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_I4C4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_I4C",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_I4Z",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_I4S4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_I4S",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_I4D",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_I4I1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_I4I2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_I4I4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_I4J",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_JC4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_DOT_JC",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_DOT_JZ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Complex_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_DOT_JS4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_DOT_JS",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_DOT_JD",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Real_16,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_DOT_JI1",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_DOT_JI2",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_DOT_JI4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_DOT_JJ",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Integer_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_DOT_L1L1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_1,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_DOT_L1L2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_2,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_DOT_L1L4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_DOT_L1L",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_DOT_L2L1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_2,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_DOT_L2L2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_2,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_DOT_L2L4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_DOT_L2L",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_DOT_L4L1",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_DOT_L4L2",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_DOT_L4L4",9,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_4,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_DOT_L4L",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_DOT_LL1",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_DOT_LL2",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_DOT_LL4",8,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_DOT_LL",7,0,1,0,1,0,1,0,0,0,Dot_Product_Intrinsic,Logical_8,0},
      {"VECTOR_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"VECTOR_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
{"DPROD",5,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DPROD",5,1,1,0,0,0,0,0,0,0,Dprod_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"DREAL",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DREAL",5,1,1,0,0,0,0,0,0,0,Dreal_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
{"DSHIFTL",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DSHIFTL",7,1,1,0,0,0,0,0,1,0,Dshiftl_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSHIFTR",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"DSHIFTR",7,1,1,0,0,0,0,0,1,0,Dshiftr_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSIGN",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DSIGN",5,1,1,0,0,0,0,0,0,0,Dsign_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DSIN",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DSIN",4,1,1,0,0,0,0,0,0,0,Dsin_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DSIND",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DSIND",5,1,1,0,0,0,0,0,0,0,Dsind_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DSINH",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DSINH",5,1,1,0,0,0,0,0,0,0,Dsinh_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
{"DSM_CHUNKSIZE",13,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_chunksize",13,0,1,0,1,0,0,0,0,0,DSM_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_DISTRIBUTION_BLOCK",22,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_block",22,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Block_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_DISTRIBUTION_CYCLIC",23,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_cyclic",23,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Cyclic_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_DISTRIBUTION_STAR",21,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_distribution_star",21,0,1,0,1,0,0,0,0,0,
				DSM_Distribution_Star_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_ISDISTRIBUTED",17,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
   {"dsm_isdistributed",17,0,1,0,1,0,0,0,0,0,
				DSM_Isdistributed_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
{"DSM_ISRESHAPED",14,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
   {"dsm_isreshaped",14,0,1,0,1,0,0,0,0,0,DSM_Isreshaped_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
{"DSM_NUMCHUNKS",13,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numchunks",14,0,1,0,1,0,0,0,0,0,DSM_Numchunks_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_NUMTHREADS",14,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_numthreads",14,0,1,0,1,0,0,0,0,0,DSM_Numthreads_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_REM_CHUNKSIZE",17,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_rem_chunksize",17,0,1,0,1,0,0,0,0,0,
					DSM_Rem_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_THIS_CHUNKSIZE",18,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_chunksize",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Chunksize_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_THIS_STARTINGINDEX",22,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_startingindex",22,0,1,0,1,0,0,0,0,0,
				DSM_This_Startingindex_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"DSM_THIS_THREADNUM",18,0,1,0,1,0,0,1,0,7,0,0,TRADITIONAL_FAMILY},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"dsm_this_threadnum",18,0,1,0,1,0,0,0,0,0,
				DSM_This_Threadnum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,0,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"INDEX",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# endif
{"DSQRT",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DSQRT",5,1,1,0,0,0,0,0,0,0,Dsqrt_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DTAN",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DTAN",4,1,1,0,0,0,0,0,0,0,Dtan_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DTAND",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"DTAND",5,1,1,0,0,0,0,0,0,0,Dtand_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"DTANH",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"DTANH",5,1,1,0,0,0,0,0,0,0,Dtanh_Intrinsic,Real_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
# ifdef KEY /* Bug 3018 */
{"DTIME",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_dtime",13,0,1,0,1,0,0,0,1,0,Dtime_Intrinsic,Real_4,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,S4_MASK,0},
{"DTIME" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_subr_dtime",18,0,0,0,1,0,0,0,1,0,Dtime_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,S4_MASK,0},
      {"RESULT",6,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
# endif /* KEY Bug 3018 */
{"ENABLE_IEEE_INTERRUPT",21,1,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"ENABLE_IEEE_INTERRUPT",21,1,0,0,0,0,0,0,0,0,
                                     Enable_Ieee_Interrupt_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"INTERRUPT",9,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"INTERRUPT",9,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"EOSHIFT",7,0,1,0,0,0,0,1,0,13,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,1,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,2,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,3,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,4,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,5,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,6,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,7,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,2,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,3,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,2,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,4,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,3,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,5,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,4,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,6,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,5,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_EOSHIFT",8,0,1,0,1,0,1,0,0,7,Eoshift_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SHIFT",5,0,0,0,0,0,1,0,0,6,0,INTEGER_MASK,0},
      {"BOUNDARY",8,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"EPSILON",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"EPSILON",7,1,1,0,0,0,0,0,0,0,Epsilon_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"EQV",3,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"EQV",3,1,1,0,0,0,0,0,1,0,Eqv_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
#ifdef KEY /* Bug 1324 */
{"ERF",3,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ERF",3,1,1,0,0,0,0,0,1,0,Erf_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"ERFC",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ERFC",4,1,1,0,0,0,0,0,1,0,Erfc_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
#endif /* KEY Bug 1324 */
# ifdef KEY /* Bug 3018 */
{"ETIME",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_etime",13,0,1,0,1,0,0,0,1,0,Etime_Intrinsic,Real_4,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,S4_MASK,0},
{"ETIME" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_subr_etime",18,0,0,0,1,0,0,0,1,0,Etime_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,S4_MASK,0},
      {"RESULT",6,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
# endif /* KEY Bug 3018 */
#ifdef KEY /* Bug 1683 */
{"EXIT",4,0,0,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"exit_byte",9,0,0,0,1,0,0,0,0,0,Exit_Intrinsic,0,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#endif /* KEY Bug 1683 */
{"EXP",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"EXP",3,1,1,0,0,0,0,0,0,0,Exp_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_COMPLEX_MASK,0},
{"EXPONENT",8,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"EXPONENT",8,1,1,0,0,0,0,0,0,0,Exponent_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"FCD",3,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"FCD",3,1,1,0,0,0,0,0,1,0,Fcd_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_POINTER_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# ifdef KEY /* Bug 1683 */
{"FDATE",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_fdate",13,0,1,0,1,0,0,0,1,0,Fdate_Intrinsic,Character_1,0},
{"FDATE" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_fdate",18,0,0,0,1,0,0,0,1,0,Fdate_Intrinsic,0,0},
      {"DATE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"FETCH_AND_ADD",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"FETCH_AND_ADD",13,1,1,0,0,0,0,0,1,0,Fetch_And_Add_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"FETCH_AND_ADD",13,1,1,0,0,0,0,0,1,0,Fetch_And_Add_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FETCH_AND_AND",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"FETCH_AND_AND",13,1,1,0,0,0,0,0,1,0,Fetch_And_And_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"FETCH_AND_AND",13,1,1,0,0,0,0,0,1,0,Fetch_And_And_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FETCH_AND_NAND",14,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"FETCH_AND_NAND",14,1,1,0,0,0,0,0,1,0,Fetch_And_Nand_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"FETCH_AND_NAND",14,1,1,0,0,0,0,0,1,0,Fetch_And_Nand_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FETCH_AND_OR",12,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"FETCH_AND_OR",12,1,1,0,0,0,0,0,1,0,Fetch_And_Or_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"FETCH_AND_OR",12,1,1,0,0,0,0,0,1,0,Fetch_And_Or_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FETCH_AND_SUB",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"FETCH_AND_SUB",13,1,1,0,0,0,0,0,1,0,Fetch_And_Sub_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"FETCH_AND_SUB",13,1,1,0,0,0,0,0,1,0,Fetch_And_Sub_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FETCH_AND_XOR",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"FETCH_AND_XOR",13,1,1,0,0,0,0,0,1,0,Fetch_And_Xor_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"FETCH_AND_XOR",13,1,1,0,0,0,0,0,1,0,Fetch_And_Xor_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
#ifdef KEY /* Bug 1683 */
{"FGET",4,0,1,0,0,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_fget",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FGET" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_fget",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FGETC",5,0,1,0,0,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"__fgetc_f90",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FGETC" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"__fgetc_f90",11,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"FLOAT",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"FLOAT",5,1,1,0,0,0,0,0,0,0,Float_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"FLOATI",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"FLOATI",6,1,1,0,0,0,0,0,0,0,Floati_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"FLOATJ",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"FLOATJ",6,1,1,0,0,0,0,0,0,0,Floatj_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"FLOATK",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"FLOATK",6,1,1,0,0,0,0,0,0,0,Floatk_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FLOOR",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"FLOOR",5,1,1,0,0,0,0,0,0,0,Floor_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"FLUSH",5,0,0,0,0,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_flush",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,1,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
#ifdef KEY
{"FNUM",4,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_fnum",14,0,1,0,1,0,0,0,1,0,Fnum_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
#endif
#ifdef KEY /* Bug 1683 */
{"FPUT",4,0,1,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_fput",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FPUT" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_fput",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FPUTC",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"__fputc_f90",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FPUTC" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"__fputc_f90",11,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK|I8_MASK,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"FP_CLASS",8,1,1,0,1,1,0,1,0,6,0,0,TRADITIONAL_FAMILY},
#else
{"FP_CLASS",8,1,1,0,1,1,0,1,0,3,0,0,0},
#endif /* KEY Bug 1683 */
   {"_FP_CLASS_I4_H",14,1,1,0,1,0,0,0,0,0,Fp_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
   {"_FP_CLASS_I8_H",14,1,1,0,1,0,0,0,0,0,Fp_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
   {"_FP_CLASS_I4_R",14,1,1,0,1,0,0,0,0,0,Fp_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
   {"_FP_CLASS_I8_R",14,1,1,0,1,0,0,0,0,0,Fp_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
   {"_FP_CLASS_I4_D",14,1,1,0,1,0,0,0,0,0,Fp_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
   {"_FP_CLASS_I8_D",14,1,1,0,1,0,0,0,0,0,Fp_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"FRACTION",8,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"FRACTION",8,1,1,0,0,0,0,0,0,0,Fraction_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
{"FREE",4,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"FREE",4,1,0,0,0,0,0,0,1,0,Free_Intrinsic,0,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,INTEGER_POINTER_MASK,0},
# else
{"FREE",4,1,0,0,1,0,0,1,0,2,0,0,PGI_FAMILY},
   {"_FREE_I4",8,1,0,0,1,0,0,0,1,0,Free_Intrinsic,0,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"_FREE_I8",8,1,0,0,1,0,0,0,1,0,Free_Intrinsic,0,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,(I8_MASK|P_MASK),0},
# endif
#ifdef KEY /* Bug 1683 */
{"FSEEK",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"__fseek_f90",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"OFFSET",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"WHENCE",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"FSEEK" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"__fseek_f90",11,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"OFFSET",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"WHENCE",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"FSEEK" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"fseekf90_4_8_4_",15,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"OFFSET",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"WHENCE",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"FSTAT",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_fstat",13,0,1,0,1,0,0,0,1,0,Fstat_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"SARRAY",6,0,0,0,0,0,0,0,0,1,0,INTEGER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"FSTAT" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_fstat",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"SARRAY",6,0,0,0,0,0,0,0,0,1,0,INTEGER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#endif /* KEY Bug 1683 */
#ifdef KEY /* Bug 1683 */
{"FTELL",5,0,1,0,1,0,0,1,0,2,0,0,PGI_FAMILY|G77_FAMILY},
   {"ftellf90_8_4_",13,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_8,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"ftellf90_8_",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_8,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"FTELL" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,3,0,0,G77_FAMILY},
   {"pathf90_ftell4",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"OFFSET",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"pathf90_ftell48",15,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"OFFSET",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
   {"pathf90_ftell8",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"OFFSET",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"GERROR",6,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_gerror_0",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"MESSAGE",7,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"GETARG",6,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getarg",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"VALUE",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"GETCWD",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getcwd",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"GETCWD" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_getcwd",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"GETENV", 6,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getenv",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"VALUE",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"GETGID",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getgid",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
{"GETLOG" ,6,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getlog",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"LOGIN",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"GETPID",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getpid",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
{"GETUID",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_getuid",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
#endif /* KEY Bug 1683 */
{"GETPOS",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"GETPOS",6,1,1,0,0,0,0,0,1,0,Getpos_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"GET_COMMAND",11,0,0,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY},
   {"_Get_command",12,0,0,0,1,0,0,0,0,0,Pathf90_Intrinsic,0,0},
      {"COMMAND",7,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
      {"LENGTH",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"GET_COMMAND_ARGUMENT",20,0,0,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY},
   {"_Get_command_argument",21,0,0,0,1,0,0,0,0,0,Pathf90_Intrinsic,0,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"VALUE",5,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
      {"LENGTH",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"GET_ENVIRONMENT_VARIABLE",24,0,0,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY},
   {"_Get_environment_variable",25,0,0,0,1,0,0,0,0,0,Pathf90_Intrinsic,0,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"VALUE",5,0,0,0,0,1,0,0,0,0,0,CHARACTER_MASK,0},
      {"LENGTH",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
      {"TRIM_NAME",9,0,0,0,0,1,0,0,0,0,0,L4_MASK,0},
{"GET_IEEE_EXCEPTIONS",19,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"GET_IEEE_EXCEPTIONS",19,0,0,0,0,0,0,0,0,0,Get_Ieee_Exceptions_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"GET_IEEE_INTERRUPTS",19,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"GET_IEEE_INTERRUPTS",19,0,0,0,0,0,0,0,0,0,Get_Ieee_Interrupts_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"GET_IEEE_ROUNDING_MODE",22,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"GET_IEEE_ROUNDING_MODE",22,0,0,0,0,0,0,0,0,0,
                                           Get_Ieee_Rounding_Mode_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"GET_IEEE_STATUS",15,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"GET_IEEE_STATUS",15,0,0,0,0,0,0,0,0,0,Get_Ieee_Status_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
#ifdef KEY /* Bug 1683 */
{"GMTIME",6,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_gmtime",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"STIME",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
{"HOSTNM",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_hostnm",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"HOSTNM" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_hostnm",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"HUGE",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"HUGE",4,1,1,0,0,0,0,0,0,0,Huge_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
{"IABS",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IABS",4,1,1,0,0,0,0,0,0,0,Iabs_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# ifdef _NEED_AT_SIGN_INTRINSICS
{"IABS@",5,1,1,1,0,0,0,1,0,1,0,0,0},
   {"IABS@",5,1,1,0,0,0,0,0,0,0,Iabs_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# endif
{"IACHAR",6,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IACHAR",6,1,1,0,0,0,0,0,0,0,Iachar_Intrinsic,Integer_4,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"IAND",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IAND",4,1,1,0,0,0,0,0,0,0,Iand_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
# ifdef _NEED_AT_SIGN_INTRINSICS
{"IAND@",5,1,1,0,0,0,0,1,0,1,0,0,0},
   {"IAND@",5,1,1,0,0,0,0,0,0,0,Iand_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
# endif
#ifdef KEY /* Bug 1683 */
{"IARGC",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_iargc",13,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
#endif /* KEY Bug 1683 */
{"IBCHNG",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IBCHNG",6,1,1,0,0,0,0,0,0,0,Ibchng_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IBCLR",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IBCLR",5,1,1,0,0,0,0,0,0,0,Ibclr_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IBITS",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IBITS",5,1,1,0,0,0,0,0,0,0,Ibits_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IBSET",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IBSET",5,1,1,0,0,0,0,0,0,0,Ibset_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ICHAR",5,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ICHAR",5,1,1,0,0,0,0,0,0,0,Ichar_Intrinsic,Integer_4,0},
      {"C",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"IDATE",5,0,0,0,1,0,0,1,0,8,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"_IDATE_I1",9,0,0,0,1,0,0,0,0,0,Idate_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I1_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I1_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,I1_MASK,0},
   {"_IDATE_I2",9,0,0,0,1,0,0,0,0,0,Idate_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
   {"_IDATE_I4",9,0,0,0,1,0,0,0,0,0,Idate_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"_IDATE_I8",9,0,0,0,1,0,0,0,0,0,Idate_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"K",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
#ifdef KEY /* Bug 1683 */
   {"_IDATE_A1",9,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I1_MASK,0},
   {"_IDATE_A2",9,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I2_MASK,0},
   {"_IDATE_A4",9,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
   {"_IDATE_A8",9,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I8_MASK,0},
#endif /* KEY Bug 1683 */
{"IDIM",4,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IDIM",4,1,1,0,0,0,0,0,0,0,Idim_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IDINT",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IDINT",5,1,1,0,0,0,0,0,0,0,Idint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"IDNINT",6,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IDNINT",6,1,1,0,0,0,0,0,0,0,Idnint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"IEEE_BINARY_SCALE",17,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_BINARY_SCALE",17,1,1,0,0,0,0,0,0,0,Ieee_Binary_Scale_Intrinsic,0,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"N",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IEEE_CLASS",10,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_CLASS",10,1,1,0,0,0,0,0,0,0,Ieee_Class_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEEE_COPY_SIGN",14,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_COPY_SIGN",14,1,1,0,0,0,0,0,0,0,Ieee_Copy_Sign_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEEE_EXPONENT",13,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_EXPONENT",13,1,1,0,0,0,0,0,0,0,Ieee_Exponent_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,INTEGER_REAL_MASK,0},
{"IEEE_FINITE",11,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_FINITE",11,1,1,0,0,0,0,0,0,0,Ieee_Finite_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEEE_INT",8,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_INT",8,1,1,0,0,0,0,0,0,0,Ieee_Int_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,INTEGER_REAL_MASK,0},
{"IEEE_IS_NAN",11,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_IS_NAN",11,1,1,0,0,0,0,0,0,0,Ieee_Is_Nan_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEEE_NEXT_AFTER",15,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_NEXT_AFTER",15,1,1,0,0,0,0,0,0,0,Ieee_Next_After_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEEE_REAL",9,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_REAL",9,1,1,0,0,0,0,0,0,0,Ieee_Real_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
{"IEEE_REMAINDER",14,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_REMAINDER",14,1,1,0,0,0,0,0,0,0,Ieee_Remainder_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEEE_UNORDERED",14,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IEEE_UNORDERED",14,1,1,0,0,0,0,0,0,0,Ieee_Unordered_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IEOR",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IEOR",4,1,1,0,0,0,0,0,0,0,Ieor_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
#ifdef KEY /* Bug 1683 */
{"IERRNO",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_ierrno_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
#endif /* KEY Bug 1683 */
{"IFIX",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IFIX",4,1,1,0,0,0,0,0,0,0,Ifix_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IIABS",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIABS",5,1,1,0,0,0,0,0,0,0,Iiabs_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"IIAND",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIAND",5,1,1,0,0,0,0,0,0,0,Iiand_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"IIBCHNG",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IIBCHNG",7,1,1,0,0,0,0,0,0,0,Iibchng_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IIBCLR",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIBCLR",6,1,1,0,0,0,0,0,0,0,Iibclr_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IIBITS",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIBITS",6,1,1,0,0,0,0,0,0,0,Iibits_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IIBSET",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIBSET",6,1,1,0,0,0,0,0,0,0,Iibset_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IIDIM",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIDIM",5,1,1,0,0,0,0,0,0,0,Iidim_Intrinsic,Integer_2,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"IIDINT",6,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIDINT",6,1,1,0,0,0,0,0,0,0,Iidint_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"IIEOR",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIEOR",5,1,1,0,0,0,0,0,0,0,Iieor_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"IIFIX",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIFIX",5,1,1,0,0,0,0,0,0,0,Iifix_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IINT",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IINT",4,1,1,0,0,0,0,0,0,0,Iint_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
{"IIOR",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IIOR",4,1,1,0,0,0,0,0,0,0,Iior_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"IIQINT",6,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IIQINT",6,1,1,0,0,0,0,0,0,0,Iiqint_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"IISHA",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IISHA",5,1,1,0,0,0,0,0,0,0,Iisha_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IISHC",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IISHC",5,1,1,0,0,0,0,0,0,0,Iishc_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IISHFT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IISHFT",6,1,1,0,0,0,0,0,0,0,Iishft_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IISHFTC",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IISHFTC",7,1,1,0,0,0,0,0,0,0,Iishftc_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"SIZE",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"IISHL",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IISHL",5,1,1,0,0,0,0,0,0,0,Iishl_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IISIGN",6,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IISIGN",6,1,1,0,0,0,0,0,0,0,Iisign_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"ILEN",4,1,1,1,0,0,0,1,0,4,0,0,TRADITIONAL_FAMILY},
   {"_ILEN_I1_I1",11,1,1,0,1,0,0,0,0,0,Ilen_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I1_MASK,0},
   {"_ILEN_I2_I2",11,1,1,0,1,0,0,0,0,0,Ilen_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
   {"_ILEN_I4_I4",11,1,1,0,1,0,0,0,0,0,Ilen_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"_ILEN_I8_I8",11,1,1,0,1,0,0,0,0,0,Ilen_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"IMAG",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"IMAG",4,1,1,0,0,0,0,0,0,0,Imag_Intrinsic,0,0},
      {"Z",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
#ifdef KEY /* Bug 1683 */
{"IMAGPART",8,0,1,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"IMAG",4,1,1,0,0,0,0,0,0,0,Imag_Intrinsic,0,0},
      {"Z",1,0,0,0,0,0,0,0,0,0,0,COMPLEX_MASK,0},
#endif /* KEY Bug 1683 */
{"IMOD",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"IMOD",4,1,1,0,0,0,0,0,0,0,Imod_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"IMVBITS",7,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IMVBITS",7,1,0,0,0,0,0,0,0,0,Imvbits_Intrinsic,0,0},
      {"FROM",4,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"FROMPOS",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"TO",2,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
      {"TOPOS",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"INDEX",5,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"INDEX",5,1,1,0,0,0,0,0,0,0,Index_Intrinsic,Integer_4,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SUBSTRING",9,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"BACK",4,0,0,0,0,1,0,0,0,0,0,LOGICAL_MASK,0},
{"ININT",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"ININT",5,1,1,0,0,0,0,0,0,0,Inint_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"INOT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"INOT",4,1,1,0,0,0,0,0,0,0,Inot_Intrinsic,Integer_2,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"INT",3,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"INT",3,1,1,0,0,0,0,0,0,0,Int_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"INT2",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"INT2",4,1,1,0,0,0,0,0,0,0,Int2_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
{"INT4",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"INT4",4,1,1,0,0,0,0,0,0,0,Int4_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
{"INT8",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"INT8",4,1,1,0,0,0,0,0,0,0,Int8_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
# if defined(_TARGET32_OR_64)
{"INT_MULT_UPPER",14,1,1,0,0,1,0,1,0,3,0,0,TRADITIONAL_FAMILY},
   {"INT_MULT_UPPER",14,1,1,0,0,0,0,0,1,0,Int_Mult_Upper_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"INT_MULT_UPPER",14,1,1,0,0,0,0,0,1,0,Int_Mult_Upper_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
   {"INT_MULT_UPPER",14,1,1,0,0,0,0,0,1,0,Int_Mult_Upper_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,TYPELESS_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,TYPELESS_MASK,0},
# else
{"INT_MULT_UPPER",14,1,1,0,0,1,0,1,0,2,0,0,0},
   {"INT_MULT_UPPER",14,1,1,0,0,0,0,0,1,0,Int_Mult_Upper_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
   {"INT_MULT_UPPER",14,1,1,0,0,0,0,0,1,0,Int_Mult_Upper_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,TYPELESS_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,TYPELESS_MASK,0},
# endif
{"IOR",3,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"IOR",3,1,1,0,0,0,0,0,0,0,Ior_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"IQINT",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IQINT",5,1,1,0,0,0,0,0,0,0,Iqint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"IQNINT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IQNINT",6,1,1,0,0,0,0,0,0,0,Iqnint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
#ifdef KEY /* Bug 1683 */
{"IRAND",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_irand_0",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"FLAG",4,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"IRTC",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"IRTC",4,1,1,0,0,0,0,0,1,0,Irtc_Intrinsic,0,0},
#ifdef KEY /* Bug 1683 */
{"ISATTY",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"__isatty_f90",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Logical_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"ISHA",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"ISHA",4,1,1,0,0,0,0,0,0,0,Isha_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ISHC",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"ISHC",4,1,1,0,0,0,0,0,0,0,Ishc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ISHFT",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ISHFT",5,1,1,0,0,0,0,0,0,0,Ishft_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ISHFTC",6,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ISHFTC",6,1,1,0,0,0,0,0,0,0,Ishftc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"SIZE",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"ISHL",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"ISHL",4,1,1,0,0,0,0,0,0,0,Ishl_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ISIGN",5,1,1,1,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"ISIGN",5,1,1,0,0,0,0,0,0,0,Isign_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"ISNAN",5,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"ISNAN",5,1,1,0,0,0,0,0,0,0,Isnan_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"IS_IOSTAT_END",13,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY},
   {"_Is_iostat_end",14,0,1,0,1,0,0,0,0,0,Pathf90_Intrinsic,Logical_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"IS_IOSTAT_EOR",13,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY},
   {"_Is_iostat_eor",14,0,1,0,1,0,0,0,0,0,Pathf90_Intrinsic,Logical_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"ITIME",5,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_itime",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"JDATE",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_F90_JDATE_",11,0,1,0,1,0,0,0,1,0,Jdate_Intrinsic,0,0},
# else
   {"_JDATE",6,0,1,0,1,0,0,0,1,0,Jdate_Intrinsic,TYPELESS_DEFAULT_TYPE,0},
# endif
{"JIABS",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIABS",5,1,1,0,0,0,0,0,0,0,Jiabs_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JIAND",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIAND",5,1,1,0,0,0,0,0,0,0,Jiand_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JIBCHNG",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JIBCHNG",7,1,1,0,0,0,0,0,0,0,Jibchng_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JIBCLR",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIBCLR",6,1,1,0,0,0,0,0,0,0,Jibclr_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JIBITS",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIBITS",6,1,1,0,0,0,0,0,0,0,Jibits_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JIBSET",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIBSET",6,1,1,0,0,0,0,0,0,0,Jibset_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JIDIM",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIDIM",5,1,1,0,0,0,0,0,0,0,Jidim_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JIDINT",6,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIDINT",6,1,1,0,0,0,0,0,0,0,Jidint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"JIEOR",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIEOR",5,1,1,0,0,0,0,0,0,0,Jieor_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JIFIX",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIFIX",5,1,1,0,0,0,0,0,0,0,Jifix_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"JINT",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JINT",4,1,1,0,0,0,0,0,0,0,Jint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
{"JIOR",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JIOR",4,1,1,0,0,0,0,0,0,0,Jior_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JIQINT",6,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JIQINT",6,1,1,0,0,0,0,0,0,0,Jiqint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"JISHA",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JISHA",5,1,1,0,0,0,0,0,0,0,Jisha_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JISHC",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JISHC",5,1,1,0,0,0,0,0,0,0,Jishc_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JISHFT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JISHFT",6,1,1,0,0,0,0,0,0,0,Jishft_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JISHFTC",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JISHFTC",7,1,1,0,0,0,0,0,0,0,Jishftc_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"SIZE",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"JISHL",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JISHL",5,1,1,0,0,0,0,0,0,0,Jishl_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JISIGN",6,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JISIGN",6,1,1,0,0,0,0,0,0,0,Jisign_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JMOD",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JMOD",4,1,1,0,0,0,0,0,0,0,Jmod_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"JMVBITS",7,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JMVBITS",7,1,0,0,0,0,0,0,0,0,Jmvbits_Intrinsic,0,0},
      {"FROM",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"FROMPOS",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"TO",2,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"TOPOS",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"JNINT",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"JNINT",5,1,1,0,0,0,0,0,0,0,Jnint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"JNOT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"JNOT",4,1,1,0,0,0,0,0,0,0,Jnot_Intrinsic,Integer_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"KIABS",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIABS",5,1,1,0,0,0,0,0,0,0,Kiabs_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KIAND",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIAND",5,1,1,0,0,0,0,0,0,0,Kiand_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KIBCHNG",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KIBCHNG",7,1,1,0,0,0,0,0,0,0,Kibchng_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KIBCLR",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIBCLR",6,1,1,0,0,0,0,0,0,0,Kibclr_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KIBITS",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIBITS",6,1,1,0,0,0,0,0,0,0,Kibits_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KIBSET",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIBSET",6,1,1,0,0,0,0,0,0,0,Kibset_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"POS",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KIDIM",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIDIM",5,1,1,0,0,0,0,0,0,0,Kidim_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KIDINT",6,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KIDINT",6,1,1,0,0,0,0,0,0,0,Kidint_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
{"KIEOR",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KIEOR",5,1,1,0,0,0,0,0,0,0,Kieor_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KIFIX",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIFIX",5,1,1,0,0,0,0,0,0,0,Kifix_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
#ifdef KEY /* Bug 1683 */
{"KILL",4,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"KILL",4,0,1,0,1,0,0,0,1,0,Kill_Intrinsic,Integer_4,0},
      {"PID",3,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SIG",3,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"KILL" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_subr_kill",17,0,0,0,1,0,0,0,1,0,Kill_Intrinsic,0,0},
      {"PID",3,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"SIG",3,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"KIND",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIND",4,1,1,0,0,0,0,0,0,0,Kind_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,ALL_MASK,0},
{"KINT",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KINT",4,1,1,0,0,0,0,0,0,0,Kint_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
{"KIOR",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KIOR",4,1,1,0,0,0,0,0,0,0,Kior_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KIQINT",6,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KIQINT",6,1,1,0,0,0,0,0,0,0,Kiqint_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"KISHA",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KISHA",5,1,1,0,0,0,0,0,0,0,Kisha_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KISHC",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KISHC",5,1,1,0,0,0,0,0,0,0,Kishc_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KISHFT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KISHFT",6,1,1,0,0,0,0,0,0,0,Kishft_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KISHFTC",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KISHFTC",7,1,1,0,0,0,0,0,0,0,Kishftc_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"SIZE",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"KISHL",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KISHL",5,1,1,0,0,0,0,0,0,0,Kishl_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"SHIFT",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KISIGN",6,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KISIGN",6,1,1,0,0,0,0,0,0,0,Kisign_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KMOD",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KMOD",4,1,1,0,0,0,0,0,0,0,Kmod_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"KMVBITS",7,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"KMVBITS",7,1,0,0,0,0,0,0,0,0,Kmvbits_Intrinsic,0,0},
      {"FROM",4,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"FROMPOS",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"TO",2,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"TOPOS",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"KNINT",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KNINT",5,1,1,0,0,0,0,0,0,0,Knint_Intrinsic,Integer_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"KNOT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"KNOT",4,1,1,0,0,0,0,0,0,0,Knot_Intrinsic,Integer_8,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"LBOUND",6,0,1,0,1,0,0,1,0,22,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_4",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND0_8",10,0,1,0,1,0,0,0,0,0,Lbound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_LBOUND",7,0,1,0,1,0,1,0,0,1,Lbound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
{"LEADZ",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"LEADZ",5,1,1,0,0,0,0,0,1,0,Leadz_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"LEN",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LEN",3,1,1,0,0,0,0,0,0,0,Len_Intrinsic,Integer_4,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"LENGTH",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"LENGTH",6,1,1,0,0,0,0,0,1,0,Length_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"LEN_TRIM",8,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LEN_TRIM",8,1,1,0,0,0,0,0,0,0,Len_Trim_Intrinsic,Integer_4,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"LGE",3,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LGE",3,1,1,0,0,0,0,0,0,0,Lge_Intrinsic,Character_1,0},
      {"STRING_A",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STRING_B",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"LGT",3,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LGT",3,1,1,0,0,0,0,0,0,0,Lgt_Intrinsic,Character_1,0},
      {"STRING_A",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STRING_B",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"LINK",4,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_link_0",10,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"PATH1",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"PATH2",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"LINK" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_link",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"PATH1",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"PATH2",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"LLE",3,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LLE",3,1,1,0,0,0,0,0,0,0,Lle_Intrinsic,Character_1,0},
      {"STRING_A",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STRING_B",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"LLT",3,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LLT",3,1,1,0,0,0,0,0,0,0,Llt_Intrinsic,Character_1,0},
      {"STRING_A",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STRING_B",8,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"LNBLNK",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_lnblnk",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"LOC",3,0,1,0,0,0,0,1,0,9,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,1,0,0,0,0,0,0,0,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,0,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,1,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,2,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,3,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,4,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,5,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,6,0,(ALL_MASK|CP_MASK),0},
   {"LOC",3,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,7,0,(ALL_MASK|CP_MASK),0},
# ifdef _NEED_AT_SIGN_INTRINSICS
{"LOC@",4,0,1,0,0,0,0,1,0,9,0,0,0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,1,0,0,0,0,0,0,0,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,0,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,1,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,2,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,3,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,4,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,5,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,6,0,(ALL_MASK|CP_MASK),0},
   {"LOC@",4,0,1,0,0,0,0,0,1,0,Loc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,1,0,0,7,0,(ALL_MASK|CP_MASK),0},
# endif
{"LOCK_RELEASE",12,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"LOCK_RELEASE",12,1,0,0,0,0,0,0,0,0,Lock_Release_Intrinsic,0,0},
      {"I",1,0,0,0,0,1,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"LOCK_TEST_AND_SET",17,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"LOCK_TEST_AND_SET",17,1,1,0,0,0,0,0,1,0,Lock_Test_And_Set_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"LOCK_TEST_AND_SET",17,1,1,0,0,0,0,0,1,0,Lock_Test_And_Set_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"LOG",3,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LOG",3,1,1,0,0,0,0,0,0,0,Log_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_COMPLEX_MASK,0},
{"LOG10",5,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LOG10",5,1,1,0,0,0,0,0,0,0,Log10_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"LOG2_IMAGES",11,0,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"shmem_n_pes",11,0,1,0,0,0,0,0,1,0,Log2_Images_Intrinsic,Integer_4,0},
{"LOGICAL",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"LOGICAL",7,1,1,0,0,0,0,0,0,0,Logical_Intrinsic,Logical_4,0},
      {"L",1,0,0,0,0,0,0,0,0,0,0,LOGICAL_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"LONG",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"LONG",4,1,1,0,0,0,0,0,1,0,Long_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
{"LSHIFT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"LSHIFT",6,1,1,0,0,0,0,0,1,0,Lshift_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"POSITIVE_SHIFT",14,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
#ifdef KEY /* Bug 1683 */
{"LSTAT",5,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_lstat",13,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"FILE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"LSTAT" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_lstat",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"FILE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"LTIME",5,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_ltime",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"STIME",5,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"TARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
# if defined(_NEED_AT_SIGN_INTRINSICS) || defined(_TARGET_OS_MAX) 
{"M@CLR",5,1,1,0,0,0,0,1,0,1,0,0,0},
   {"M@CLR",5,1,1,0,0,0,0,0,1,0,Mclr_Intrinsic,0,0},
{"M@LD",4,1,1,0,0,0,0,1,0,1,0,0,0},
   {"M@LD",4,1,1,0,0,0,0,0,1,0,Mld_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"M@LDMX",6,1,1,0,0,0,0,1,0,1,0,0,0},
   {"M@LDMX",6,1,1,0,0,0,0,0,1,0,Mldmx_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"M@MX",4,1,1,0,0,0,0,1,0,2,0,0,0},
   {"M@MX",4,1,1,0,0,0,0,0,1,0,Mmx_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
   {"M@MX",4,1,1,0,0,0,0,0,1,0,Mldmx_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"M@UL",4,1,1,0,0,0,0,1,0,1,0,0,0},
   {"M@UL",4,1,1,0,0,0,0,0,1,0,Mul_Intrinsic,0,0},
# endif
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
{"MALLOC",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"MALLOC",6,1,1,0,0,0,0,0,0,0,Malloc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# else
{"MALLOC",6,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_MALLOC_I4_I4",13,1,1,0,1,0,0,0,1,0,Malloc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,(I1_MASK|I2_MASK|I4_MASK),0},
   {"_MALLOC_I8_I4",13,1,1,0,1,0,0,0,1,0,Malloc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,(I1_MASK|I2_MASK|I4_MASK),0},
   {"_MALLOC_I4_I8",13,1,1,0,1,0,0,0,1,0,Malloc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,(I8_MASK|P_MASK),0},
   {"_MALLOC_I8_I8",13,1,1,0,1,0,0,0,1,0,Malloc_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,(I8_MASK|P_MASK),0},
# endif
{"MASK",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"MASK",4,1,1,0,0,0,0,0,1,0,Mask_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"MATMUL",6,0,1,0,0,0,0,1,0,348,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_MATMUL_C4C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_C4C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_C4Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_C4S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_C4S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_C4D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_C4I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_C4I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_C4I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_C4J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_CC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_CC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_CZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_CS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_CS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_CD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_CI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_CI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_CI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_CJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_DC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_DC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_DZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_DS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_DS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_DD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_DI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_DI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_DI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_DJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_I1C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_I1C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_I1Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_I1S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_I1S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_I1D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_I1I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_1,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_I1I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_I1I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_I1J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_I2C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_I2C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_I2Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_I2S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_I2S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_I2D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_I2I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_I2I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_I2I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_I2J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_I4C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_I4C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_I4Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_I4S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_I4S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_I4D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_I4I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_I4I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_I4I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_I4J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_JC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_JC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_JZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_JS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_JS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_JD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_JI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_JI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_JI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_JJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_S4C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_S4C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_S4Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_S4S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_S4S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_S4D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_S4I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_S4I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_S4I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_S4J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_SC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_SC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_SZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_SS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_SS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_SD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_SI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_SI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_SI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_SJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_ZC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_ZC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_ZZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_ZS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_ZS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_ZD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_ZI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_ZI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_ZI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_ZJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_L1L1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_1,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_L1L2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_L1L4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_L1L",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_L2L1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_L2L2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_L2L4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_L2L",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_L4L1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_L4L2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_L4L4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_L4L",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_LL1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_LL2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_LL4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_LL",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_C4C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_C4C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_C4Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_C4S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_C4S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_C4D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_C4I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_C4I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_C4I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_C4J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_CC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_CC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_CZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_CS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_CS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_CD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_CI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_CI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_CI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_CJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_DC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_DC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_DZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_DS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_DS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_DD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_DI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_DI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_DI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_DJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_I1C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_I1C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_I1Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_I1S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_I1S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_I1D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_I1I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_1,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_I1I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_I1I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_I1J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_I2C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_I2C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_I2Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_I2S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_I2S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_I2D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_I2I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_I2I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_I2I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_I2J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_I4C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_I4C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_I4Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_I4S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_I4S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_I4D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_I4I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_I4I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_I4I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_I4J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_JC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_JC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_JZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_JS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_JS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_JD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_JI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_JI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_JI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_JJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_S4C4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_S4C",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_S4Z",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_S4S4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_S4S",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_S4D",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_S4I1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_S4I2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_S4I4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_S4J",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_SC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_SC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_SZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_SS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_SS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_SD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_SI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_SI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_SI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_SJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_ZC4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_MATMUL_ZC",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_MATMUL_ZZ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_MATMUL_ZS4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MATMUL_ZS",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MATMUL_ZD",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MATMUL_ZI1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MATMUL_ZI2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MATMUL_ZI4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MATMUL_ZJ",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MATMUL_L1L1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_1,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_MATMUL_L1L2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_MATMUL_L1L4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_MATMUL_L1L",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_MATMUL_L2L1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_MATMUL_L2L2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_MATMUL_L2L4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_MATMUL_L2L",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_MATMUL_L4L1",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_MATMUL_L4L2",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_MATMUL_L4L4",12,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_MATMUL_L4L",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_MATMUL_LL1",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L1_MASK,0},
   {"_MATMUL_LL2",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L2_MASK,0},
   {"_MATMUL_LL4",11,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L4_MASK,0},
   {"_MATMUL_LL",10,0,1,0,1,0,1,0,0,1,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,1,0,L8_MASK,0},
   {"_MATMUL_C4C4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_C4C",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_C4Z",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_C4S4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_C4S",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_C4D",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_C4I1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_C4I2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_C4I4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_C4J",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_CC4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_CC",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_CZ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_CS4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_CS",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_CD",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_CI1",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_CI2",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_CI4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_CJ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_DC4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_DC",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_DZ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_DS4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_DS",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_DD",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_DI1",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_DI2",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_DI4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_DJ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_I1C4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_I1C",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_I1Z",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_I1S4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_I1S",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_I1D",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_I1I1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_1,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_I1I2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_I1I4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_I1J",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_I2C4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_I2C",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_I2Z",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_I2S4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_I2S",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_I2D",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_I2I1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_I2I2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_I2I4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_I2J",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_I4C4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_I4C",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_I4Z",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_I4S4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_I4S",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_I4D",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_I4I1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_I4I2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_I4I4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_I4J",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_JC4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_JC",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_JZ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_JS4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_JS",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_JD",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_JI1",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_JI2",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_JI4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_JJ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Integer_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_S4C4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_S4C",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_S4Z",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_S4S4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_S4S",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_S4D",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_S4I1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_S4I2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_S4I4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_S4J",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_SC4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_SC",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_SZ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_SS4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_SS",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_SD",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_SI1",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_SI2",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_SI4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_SJ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Real_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_ZC4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_MATMUL_ZC",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_MATMUL_ZZ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_MATMUL_ZS4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MATMUL_ZS",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MATMUL_ZD",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MATMUL_ZI1",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MATMUL_ZI2",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MATMUL_ZI4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MATMUL_ZJ",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Complex_16,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MATMUL_L1L1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_1,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_L1L2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_L1L4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_L1L",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_L2L1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_L2L2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_2,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_L2L4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_L2L",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_L4L1",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_L4L2",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_L4L4",12,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_4,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_L4L",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
   {"_MATMUL_LL1",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L1_MASK,0},
   {"_MATMUL_LL2",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L2_MASK,0},
   {"_MATMUL_LL4",11,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L4_MASK,0},
   {"_MATMUL_LL",10,0,1,0,1,0,1,0,0,2,Matmul_Intrinsic,Logical_8,0},
      {"MATRIX_A",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
      {"MATRIX_B",8,0,0,0,0,0,1,0,0,2,0,L8_MASK,0},
{"MAX",3,1,1,0,0,0,1,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MAX0",4,1,1,0,0,0,0,0,0,0,Max_Intrinsic,Integer_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"AMAX1",5,1,1,0,0,0,0,0,0,0,Max_Intrinsic,Real_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
{"MAX0",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MAX0",4,1,1,0,0,0,0,0,0,0,Max0_Intrinsic,Integer_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"MAX1",4,1,1,0,0,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MAX1",4,1,1,0,0,0,0,0,0,0,Max1_Intrinsic,Real_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
   {"MAX1",4,1,1,0,0,0,0,0,0,0,Max1_Intrinsic,Real_8,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
{"MAXEXPONENT",11,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"MAXEXPONENT",11,1,1,0,0,0,0,0,0,0,Maxexponent_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"MAXLOC",6,0,1,0,0,0,0,1,0,294,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I1",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I2",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__I4",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__J",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S4",11,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__S",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,0,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,1,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,2,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,3,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,4,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,5,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXLOC__D",10,0,1,0,1,0,1,0,0,6,Maxloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
{"MAXVAL",6,0,1,0,0,0,0,1,0,294,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I1",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I1",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I2",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I2",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__I4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__I4",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__J",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__J",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S4",12,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S4",11,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__S",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__S",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL0__D",11,0,1,0,1,0,0,0,0,0,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,1,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,2,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,3,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,4,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,5,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MAXVAL__D",10,0,1,0,1,0,1,0,0,6,Maxval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
#ifdef KEY /* Bug 1683 */
{"MCLOCK",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_mclock",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
{"MCLOCK8",7,0,1,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_mclock8",15,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_8,0},
#endif /* KEY Bug 1683 */
{"MEMORY_BARRIER",14,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"MEMORY_BARRIER",14,1,0,0,0,0,0,0,1,0, Memory_Barrier_Intrinsic,0,0},
{"MERGE",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"MERGE",5,1,1,0,0,0,0,0,0,0,Merge_Intrinsic,0,0},
      {"TSOURCE",7,0,0,0,0,0,0,0,0,0,0,ALL_MASK,0},
      {"FSOURCE",7,0,0,0,0,0,0,0,0,0,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,0,0,0,0,0,LOGICAL_MASK,0},
{"MIN",3,1,1,0,0,0,1,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MIN0",4,1,1,0,0,0,0,0,0,0,Min_Intrinsic,Integer_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"AMIN1",5,1,1,0,0,0,0,0,0,0,Min_Intrinsic,Real_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
{"MIN0",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MIN0",4,1,1,0,0,0,0,0,0,0,Min0_Intrinsic,Integer_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"MIN1",4,1,1,0,0,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MIN1",4,1,1,0,0,0,0,0,0,0,Min1_Intrinsic,Real_4,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
   {"MIN1",4,1,1,0,0,0,0,0,0,0,Min1_Intrinsic,Real_8,0},
      {"A1",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A2",2,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"A3",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A4",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A5",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A6",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A7",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A8",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A9",2,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A10",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A11",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A12",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A13",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A14",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A15",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A16",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A17",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A18",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A19",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A20",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A21",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A22",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A23",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A24",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A25",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A26",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A27",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A28",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A29",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A30",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A31",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A32",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A33",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A34",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A35",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A36",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A37",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A38",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A39",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A40",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A41",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A42",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A43",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A44",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A45",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A46",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A47",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A48",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A49",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A50",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A51",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A52",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A53",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A54",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A55",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A56",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A57",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A58",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A59",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A60",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A61",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A62",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
      {"A63",3,0,0,0,0,1,0,0,0,0,0,S8_MASK,0},
{"MINEXPONENT",11,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"MINEXPONENT",11,1,1,0,0,0,0,0,0,0,Minexponent_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"MINLOC",6,0,1,0,0,0,0,1,0,294,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I1",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I2",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__I4",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__J",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S4",11,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__S",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,0,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,1,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,2,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,3,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,4,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,5,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINLOC__D",10,0,1,0,1,0,1,0,0,6,Minloc_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
{"MINVAL",6,0,1,0,0,0,0,1,0,294,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I1",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I1",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I2",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I2",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__I4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__I4",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__J",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__J",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S4",12,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S4",11,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__S",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__S",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL0__D",11,0,1,0,1,0,0,0,0,0,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,1,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,2,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,3,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,4,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,5,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_MINVAL__D",10,0,1,0,1,0,1,0,0,6,Minval_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
{"MOD",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MOD",3,1,1,0,0,0,0,0,0,0,Mod_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
{"MODULO",6,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"MODULO",6,1,1,0,0,0,0,0,0,0,Modulo_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
{"MVBITS",6,1,0,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"MVBITS",6,1,0,0,0,0,0,0,0,0,Mvbits_Intrinsic,0,0},
      {"FROM",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"FROMPOS",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"LEN",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"TO",2,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"TOPOS",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# if defined(_TARGET_OS_MAX)
{"MY_PE",5,1,1,0,0,0,0,1,0,1,0,0,0},
   {"MY_PE",5,1,1,0,0,0,0,0,1,0,My_Pe_Intrinsic,Integer_4,0},
# endif
{"NAND_AND_FETCH",14,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"NAND_AND_FETCH",14,1,1,0,0,0,0,0,1,0,Nand_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"NAND_AND_FETCH",14,1,1,0,0,0,0,0,1,0,Nand_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"NEAREST",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"NEAREST",7,1,1,0,0,0,0,0,0,0,Nearest_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"S",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"NEQV",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"NEQV",4,1,1,0,0,0,0,0,1,0,Neqv_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"NEW_LINE",8,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"NEW_LINE",8,1,1,0,0,0,0,0,0,0,Newline_Intrinsic,Character_1,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"NINT",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"NINT",4,1,1,0,0,0,0,0,0,0,Nint_Intrinsic,Integer_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"NOT",3,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"NOT",3,1,1,0,0,0,0,0,0,0,Not_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"NULL",4,0,1,0,0,0,0,1,0,8,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,2,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,3,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,4,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,5,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,6,0,ALL_MASK,0},
   {"NULL",4,0,1,0,0,0,1,0,0,0,Null_Intrinsic,0,0},
      {"MOLD",4,0,0,0,0,1,1,0,0,7,0,ALL_MASK,0},
# if defined(_TARGET_OS_UNICOS) || defined(_TARGET_OS_MAX)
{"NUMARG",6,1,1,0,0,0,0,1,0,1,0,0,0},
   {"NUMARG",6,1,1,0,0,0,0,0,1,0,Numarg_Intrinsic,0,0},
# endif
{"NUM_IMAGES",10,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"shmem_n_pes",11,0,1,0,1,0,0,0,1,0,Num_Images_Intrinsic,Integer_4,0},
#ifdef KEY
{"OMP_DESTROY_LOCK",16,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_destroy_lock_",17,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_DESTROY_NEST_LOCK",21,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_destroy_nest_lock_",22,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
#endif
{"OMP_GET_DYNAMIC",15,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_dynamic_",16,0,1,0,1,0,0,0,1,0,Omp_Get_Dynamic_Intrinsic,0,0},
{"OMP_GET_MAX_THREADS",19,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_max_threads_",20,0,1,0,1,0,0,0,1,0,
				Omp_Get_Max_Threads_Intrinsic,0,0},
{"OMP_GET_NESTED",14,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_nested_",15,0,1,0,1,0,0,0,1,0,Omp_Get_Nested_Intrinsic,0,0},
{"OMP_GET_NUM_PROCS",17,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_num_procs_",18,0,1,0,1,0,0,0,1,0,Omp_Get_Num_Procs_Intrinsic,0,0},
{"OMP_GET_NUM_THREADS",19,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_num_threads_",20,0,1,0,1,0,0,0,1,0,
				Omp_Get_Num_Threads_Intrinsic,0,0},
{"OMP_GET_THREAD_NUM",18,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_thread_num_",19,0,1,0,1,0,0,0,1,0,Omp_Get_Thread_Num_Intrinsic,0,0},
#ifdef KEY
{"OMP_GET_WTICK",13,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_wtick_",14,0,1,0,1,0,0,0,1,0,Omp_Get_Wtick_Intrinsic,Real_8,0},
{"OMP_GET_WTIME",13,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_get_wtime_",14,0,1,0,1,0,0,0,1,0,Omp_Get_Wtime_Intrinsic,Real_8,0},
{"OMP_INIT_LOCK",13,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_init_lock_",14,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_INIT_NEST_LOCK",18,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_init_nest_lock_",19,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
#endif
{"OMP_IN_PARALLEL",15,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_in_parallel_",16,0,1,0,1,0,0,0,1,0,Omp_In_Parallel_Intrinsic,0,0},
#ifdef KEY
{"OMP_SET_DYNAMIC",15,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_set_dynamic_",16,0,0,0,1,0,0,0,0,0,0,0,0},
      {"DYNAMIC_THREADS",16,0,0,0,0,0,0,0,0,0,0,(L4_MASK|L8_MASK),0},
{"OMP_SET_LOCK",12,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_set_lock_",13,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_SET_NESTED",14,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_set_nested_",15,0,0,0,1,0,0,0,0,0,0,0,0},
      {"NESTED",6,0,0,0,0,0,0,0,0,0,0,(L4_MASK|L8_MASK),0},
{"OMP_SET_NEST_LOCK",17,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_set_nest_lock_",18,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_SET_NUM_THREADS",19,0,0,0,0,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_set_num_threads_",20,0,0,0,0,0,0,0,1,0,0,0,0},
      {"NUM_THREADS",11,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_TEST_LOCK",13,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_test_lock_",14,0,1,0,1,0,0,0,1,0,Omp_Test_Lock_Intrinsic,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_TEST_NEST_LOCK",18,0,1,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_test_nest_lock_",19,0,1,0,1,0,0,0,1,0,Omp_Test_Nest_Lock_Intrinsic,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_UNSET_LOCK",14,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_unset_lock_",15,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_UNSET_NEST_LOCK",19,0,0,0,1,0,0,1,0,1,0,0,OMP_FAMILY},
   {"omp_unset_nest_lock_",20,0,0,0,1,0,0,0,0,0,0,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
#else
{"OMP_SET_LOCK",12,0,0,0,0,0,0,1,0,1,0,0,0},
   {"OMP_SET_LOCK",12,0,0,0,0,0,0,0,1,0,Omp_Set_Lock_Intrinsic,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_TEST_LOCK",13,0,1,0,0,0,0,1,0,1,0,0,0},
   {"OMP_TEST_LOCK",13,0,1,0,0,0,0,0,1,0,Omp_Test_Lock_Intrinsic,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
{"OMP_UNSET_LOCK",14,0,0,0,0,0,0,1,0,1,0,0,0},
   {"OMP_UNSET_LOCK",14,0,0,0,0,0,0,0,1,0,Omp_Unset_Lock_Intrinsic,0,0},
      {"LOCK",4,0,0,0,0,0,0,0,0,0,0,(I4_MASK|I8_MASK),0},
#endif
{"OR",2,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 1683 */
   /* For g77 compatibility, must treat values bitwise regardless of type */
   {"OR",2,1,1,0,0,0,0,0,1,0,Ior_Intrinsic,0,0},
#else
   {"OR",2,1,1,0,0,0,0,0,1,0,Or_Intrinsic,0,0},
#endif /* KEY Bug 1683 */
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"OR_AND_FETCH",12,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"OR_AND_FETCH",12,1,1,0,0,0,0,0,1,0,Or_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"OR_AND_FETCH",12,1,1,0,0,0,0,0,1,0,Or_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"PACK",4,0,1,0,1,0,0,1,0,14,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
   {"_PACK",5,0,1,0,1,0,1,0,0,1,Pack_Intrinsic,0,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
      {"VECTOR",6,0,0,0,0,1,1,0,0,1,0,ALL_MASK,0},
#ifdef KEY /* Bug 1683 */
{"PERROR",6,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_perror_0",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"POPCNT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"POPCNT",6,1,1,0,0,0,0,0,1,0,Popcnt_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"POPPAR",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"POPPAR",6,1,1,0,0,0,0,0,1,0,Poppar_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"PRECISION",9,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"PRECISION",9,1,1,0,0,0,0,0,0,0,Precision_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_COMPLEX_MASK,0},
{"PRESENT",7,1,1,0,0,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"PRESENT",7,1,1,0,0,0,0,0,0,0,Present_Intrinsic,0,0},
      {"A",1,0,1,0,0,0,0,0,0,0,0,ALL_MASK,0},
   {"PRESENT",7,1,1,0,0,0,0,0,0,0,Present_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,ALL_MASK,0},
{"PRODUCT",8,0,1,0,0,0,0,1,0,420,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I1",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I1",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I2",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I2",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__I4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__I4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__J",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__J",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__S",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__S",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__D",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__D",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C4",10,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C4",9,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD0__C",9,0,1,0,1,0,0,0,0,0,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__C",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,0,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,1,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,2,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,3,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,4,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,5,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_PROD__Z",8,0,1,0,1,0,1,0,0,6,Product_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
{"QABS",4,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QABS",4,1,1,0,0,0,0,0,0,0,Qabs_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QACOS",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QACOS",5,1,1,0,0,0,0,0,0,0,Qacos_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QACOSD",6,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QACOSD",6,1,1,0,0,0,0,0,0,0,Qacosd_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QASIN",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QASIN",5,1,1,0,0,0,0,0,0,0,Qasin_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QASIND",6,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QASIND",6,1,1,0,0,0,0,0,0,0,Qasind_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QATAN",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QATAN",5,1,1,0,0,0,0,0,0,0,Qatan_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QATAN2",6,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QATAN2",6,1,1,0,0,0,0,0,0,0,Qatan2_Intrinsic,Real_16,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QATAN2D",7,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QATAN2D",7,1,1,0,0,0,0,0,0,0,Qatan2d_Intrinsic,Real_16,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QATAND",6,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QATAND",6,1,1,0,0,0,0,0,0,0,Qatand_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QCMPLX",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QCMPLX",6,1,1,0,0,0,0,0,0,0,Qcmplx_Intrinsic,Complex_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
      {"Y",1,0,0,0,0,1,0,0,0,0,0,IRC_MASK,0},
{"QCOS",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QCOS",4,1,1,0,0,0,0,0,1,0,Qcos_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QCOSD",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QCOSD",5,1,1,0,0,0,0,0,1,0,Qcosd_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QCOSH",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QCOSH",5,1,1,0,0,0,0,0,1,0,Qcosh_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QCOT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QCOT",4,1,1,0,0,0,0,0,1,0,Qcot_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QCOTAN",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QCOTAN",6,1,1,0,0,0,0,0,1,0,Qcot_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QDIM",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QDIM",4,1,1,0,0,0,0,0,0,0,Qdim_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QEXP",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QEXP",4,1,1,0,0,0,0,0,1,0,Qexp_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QEXT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QEXT",4,1,1,0,0,0,0,0,1,0,Qext_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
{"QFLOAT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QFLOAT",6,1,1,0,0,0,0,0,0,0,Qfloat_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"QFLOATI",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QFLOATI",7,1,1,0,0,0,0,0,0,0,Qfloati_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I2_MASK,0},
{"QFLOATJ",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QFLOATJ",7,1,1,0,0,0,0,0,0,0,Qfloatj_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"QFLOATK",7,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QFLOATK",7,1,1,0,0,0,0,0,0,0,Qfloatk_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"QIMAG",5,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QIMAG",5,1,1,0,0,0,0,0,0,0,Qimag_Intrinsic,Real_16,0},
      {"Z",1,0,0,0,0,0,0,0,0,0,0,C16_MASK,0},
{"QINT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QINT",4,1,1,0,0,0,0,0,0,0,Qint_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QLOG",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QLOG",4,1,1,0,0,0,0,0,1,0,Qlog_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QLOG10",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QLOG10",6,1,1,0,0,0,0,0,1,0,Qlog10_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QMOD",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QMOD",4,1,1,0,0,0,0,0,0,0,Qmod_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"P",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QNINT",5,1,1,1,0,0,0,1,0,3,0,0,TRADITIONAL_FAMILY},
   {"QNINT",5,1,1,0,0,0,0,0,0,0,Qnint_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QPROD",5,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QPROD",5,1,1,0,0,0,0,0,0,0,Qprod_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"Y",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"QREAL",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QREAL",5,1,1,0,0,0,0,0,0,0,Qreal_Intrinsic,Real_16,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
{"QSIGN",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_QSIGN",6,1,1,0,0,0,0,0,0,0,Qsign_Intrinsic,Real_16,0},
# else
   {"QSIGN",5,1,1,0,0,0,0,0,0,0,Qsign_Intrinsic,Real_16,0},
# endif
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QSIN",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QSIN",4,1,1,0,0,0,0,0,1,0,Qsin_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QSIND",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QSIND",5,1,1,0,0,0,0,0,1,0,Qsind_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QSINH",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QSINH",5,1,1,0,0,0,0,0,1,0,Qsinh_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QSQRT",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QSQRT",5,1,1,0,0,0,0,0,1,0,Qsqrt_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QTAN",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QTAN",4,1,1,0,0,0,0,0,1,0,Qtan_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QTAND",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QTAND",5,1,1,0,0,0,0,0,1,0,Qtand_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"QTANH",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"QTANH",5,1,1,0,0,0,0,0,1,0,Qtanh_Intrinsic,Real_16,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"RADIX",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"RADIX",5,1,1,0,0,0,0,0,0,0,Radix_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
/*
   Deleting follow table entries fixes pv 782741
{"RAN",3,0,1,0,0,0,0,1,0,8,0,0,PGI_FAMILY},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,2,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,3,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,4,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,5,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,6,0,I4_MASK,0},
   {"RAN",3,0,1,0,0,0,0,0,1,0,Ran_Intrinsic,Real_4,0},
      {"I",1,0,0,0,0,0,0,0,0,7,0,I4_MASK,0},
*/
#ifdef KEY /* Bug 1683 */
{"RAND",4,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_rand_0",10,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
      {"FLAG",4,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"RANDOM_NUMBER",13,1,0,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"RANDOM_NUMBER",13,1,0,0,0,0,0,0,0,0,Random_Number_Intrinsic,0,0},
      {"HARVEST",7,0,0,0,0,1,0,0,0,0,0,REAL_MASK,0},
{"RANDOM_SEED",11,0,0,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"RANDOM_SEED",11,0,0,0,0,0,0,0,0,0,Random_Seed_Intrinsic,0,0},
      {"SIZE",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"PUT",3,0,0,0,0,1,1,0,0,1,0,INTEGER_MASK,0},
      {"GET",3,0,0,0,0,1,1,0,0,1,0,INTEGER_MASK,0},
{"RANF",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"RANF",4,1,1,0,0,0,0,0,1,0,Ranf_Intrinsic,0,0},
{"RANGE",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"RANGE",5,1,1,0,0,0,0,0,0,0,Range_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
#ifdef KEY /* Bug 3686 */
/* Direct user access to "ranget" and "ranset" is obsolete because user-level
 * interface passes a single integer*8 value and the seed no longer fits;
 * use standard Fortran "random_seed" instead */
#else /* KEY Bug 3686 */
{"RANGET",6,1,1,0,0,0,0,1,0,1,0,0,0},
   {"RANGET",6,1,1,0,0,0,0,0,1,0,Ranget_Intrinsic,TYPELESS_DEFAULT_TYPE,0},
      {"I",1,0,0,0,0,1,0,0,0,0,0,IRT_MASK,0},
# ifdef _TARGET_OS_MAX
{"RANSET",6,1,1,0,0,0,0,1,0,1,0,0,0},
   {"RANSET",6,1,1,0,0,0,0,0,1,0,Ranset_Intrinsic,TYPELESS_DEFAULT_TYPE,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRT_MASK,0},
      {"J",1,0,0,0,0,1,0,0,0,0,0,IRT_MASK,0},
# else
{"RANSET",6,1,1,0,0,0,0,1,0,1,0,0,0},
   {"RANSET",6,1,1,0,0,0,0,0,1,0,Ranset_Intrinsic,TYPELESS_DEFAULT_TYPE,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRT_MASK,0},
# endif
#endif /* KEY Bug 3686 */
# ifdef _NEED_AT_SIGN_INTRINSICS
{"READ@SM",7,1,1,0,0,0,0,1,0,1,0,0,0},
   {"READ@SM",7,1,1,0,0,0,0,0,1,0,Readsm_Intrinsic,0,0},
# endif
{"REAL",4,1,1,0,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"REAL",5,1,1,0,0,0,0,0,0,0,Real_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"REALPART",8,0,1,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"REAL",5,1,1,0,0,0,0,0,0,0,Real_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,(IRC_MASK|TYPELESS_MASK),0},
      {"KIND",4,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
#endif /* KEY Bug 1683 */
{"REMOTE_WRITE_BARRIER",20,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"REMOTE_WRITE_BARRIER",20,1,0,0,0,0,0,0,1,0,
                               Remote_Write_Barrier_Intrinsic,0,0},
{"REM_IMAGES",10,0,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"shmem_n_pes",11,0,1,0,0,0,0,0,1,0,Rem_Images_Intrinsic,Integer_4,0},
#ifdef KEY /* Bug 1683 */
{"RENAME",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_rename",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"PATH1",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"PATH2",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"RENAME" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_rename",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"PATH1",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"PATH2",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"REPEAT",6,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_F90_REPEAT",11,0,1,0,1,0,1,0,0,0,Repeat_Intrinsic,0,0},
# else
   {"_REPEAT",7,0,1,0,1,0,1,0,0,0,Repeat_Intrinsic,0,0},
# endif
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"RESHAPE",7,0,1,0,1,0,0,1,0,7,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
   {"_RESHAPE",8,0,1,0,1,0,1,0,0,1,Reshape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SHAPE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
      {"PAD",3,0,0,0,0,1,1,0,0,0,0,ALL_MASK,0},
      {"ORDER",5,0,0,0,0,1,1,0,0,0,0,INTEGER_MASK,0},
{"RRSPACING",9,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"RRSPACING",9,1,1,0,0,0,0,0,0,0,Rrspacing_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"RSHIFT",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 1683 */
   /* Although the g77 documentation says that "rshift" isn't precisely
    * defined, experiment shows that it performs an arithmetic shift. */
   {"RSHIFT",6,1,1,0,0,0,0,0,1,0,Shifta_Intrinsic,0,0},
#else
   {"RSHIFT",6,1,1,0,0,0,0,0,1,0,Rshift_Intrinsic,0,0},
#endif /* KEY Bug 1683 */
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"NEGATIVE_SHIFT",14,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"RTC",3,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"RTC",3,1,1,0,0,0,0,0,1,0,Rtc_Intrinsic,0,0},
{"SCALE",5,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"SCALE",5,1,1,0,0,0,0,0,0,0,Scale_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
# ifdef _TARGET_OS_MAX
{"SCAN",4,1,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_SCAN_4",7,1,1,0,1,0,0,0,0,0,Scan_Intrinsic,Integer_4,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SET",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"BACK",4,0,0,0,0,1,0,0,0,0,0,LOGICAL_MASK,0},
   {"_SCAN_8",7,1,1,0,1,0,0,0,0,0,Scan_Intrinsic,Integer_8,0},
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SET",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"BACK",4,0,0,0,0,1,0,0,0,0,0,LOGICAL_MASK,0},
# else
{"SCAN",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
#ifdef KEY /* Bug 2611 */
   /* Imitate behavior of "index" with respect to result type */
   {"SCAN",4,1,1,0,0,0,0,0,0,0,Scan_Intrinsic,Integer_4,0},
#else /* KEY Bug 2611 */
   {"SCAN",4,1,1,0,0,0,0,0,0,0,Scan_Intrinsic,0,0},
#endif /* KEY Bug 2611 */
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SET",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"BACK",4,0,0,0,0,1,0,0,0,0,0,LOGICAL_MASK,0},
# endif
#ifdef KEY /* Bug 1683 */
{"SECNDS",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"secnds_vms",10,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"T",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
{"SECOND",6,0,1,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_second",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_4,0},
      {"SECONDS",7,0,0,0,0,1,0,0,0,0,0,S4_MASK,0},
{"SECOND" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_second",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"SECONDS",7,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
#endif /* KEY Bug 1683 */
{"SELECTED_INT_KIND",17,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"SELECTED_INT_KIND",17,0,1,0,0,0,0,0,0,0,SIK_Intrinsic,0,0},
      {"R",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"SELECTED_REAL_KIND",18,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_SELECTED_REAL_KIND_4",21,0,1,0,1,0,0,0,0,0,SRK_Intrinsic,0,0},
      {"P",1,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"R",1,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
   {"_SELECTED_REAL_KIND_8",21,0,1,0,1,0,0,0,0,0,SRK_Intrinsic,0,0},
      {"P",1,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
      {"R",1,0,0,0,0,1,0,0,0,0,0,INTEGER_MASK,0},
{"SETBUF",6,0,1,0,0,0,0,1,0,1,0,0,0},
   {"_Setbuf",7,0,1,0,1,0,0,0,0,0,Pathf90_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"BUF",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
{"SETLINEBUF",10,0,1,0,0,0,0,1,0,1,0,0,0},
   {"_Setlinebuf",11,0,1,0,1,0,0,0,0,0,Pathf90_Intrinsic,Integer_4,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"SET_EXPONENT",12,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"SET_EXPONENT",12,1,1,0,0,0,0,0,0,0,Set_Exponent_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"SET_IEEE_EXCEPTION",18,1,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SET_IEEE_EXCEPTION",18,1,0,0,0,0,0,0,0,0,Set_Ieee_Exception_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"EXCEPTION",9,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"EXCEPTION",9,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"SET_IEEE_EXCEPTIONS",19,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SET_IEEE_EXCEPTIONS",19,0,0,0,0,0,0,0,0,0,Set_Ieee_Exceptions_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"SET_IEEE_INTERRUPTS",19,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SET_IEEE_INTERRUPTS",19,0,0,0,0,0,0,0,0,0,Set_Ieee_Interrupts_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"SET_IEEE_ROUNDING_MODE",22,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SET_IEEE_ROUNDING_MODE",22,0,0,0,0,0,0,0,0,0,
                                         Set_Ieee_Rounding_Mode_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"SET_IEEE_STATUS",15,0,0,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SET_IEEE_STATUS",15,0,0,0,0,0,0,0,0,0,Set_Ieee_Status_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"STATUS",6,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"SHAPE",5,0,1,0,1,0,0,1,0,8,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_SHAPE",6,0,1,0,1,0,1,0,0,1,Shape_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
{"SHIFT",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"SHIFT",5,1,1,0,0,0,0,0,1,0,Shift_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"SHIFTA",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SHIFTA",6,1,1,0,0,0,0,0,1,0,Shifta_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"SHIFTL",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SHIFTL",6,1,1,0,0,0,0,0,1,0,Shiftl_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"SHIFTR",6,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SHIFTR",6,1,1,0,0,0,0,0,1,0,Shiftr_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
{"SHORT",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"SHORT",5,1,1,0,0,0,0,0,1,0,Short_Intrinsic,Integer_2,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,IRC_MASK,0},
#ifdef KEY /* Bug 3410 */
{"SIGN",4,1,1,1,0,0,1,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"SIGN",4,1,1,0,0,0,0,0,0,0,Sign_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,INTEGER_REAL_MASK,0},
#else
{"SIGN",4,1,1,1,0,0,1,1,0,4,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"SIGN",4,1,1,0,0,0,0,0,0,0,Sign_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"SIGN",4,1,1,0,0,0,0,0,0,0,Sign_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,S4_MASK,0},
   {"SIGN",4,1,1,0,0,0,0,0,0,0,Sign_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
# endif
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_QSIGN",6,1,1,0,0,0,0,0,0,0,Sign_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
# else
   {"SIGN",4,1,1,0,0,0,0,0,0,0,Sign_Intrinsic,0,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
      {"B",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
#endif /* KEY Bug 3410 */
#ifdef KEY /* Bug 1603 */
{"SIGNAL",6,0,1,0,1,0,0,1,0,4,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_signal",14,0,1,0,1,0,0,0,1,0,Signal_Intrinsic,Integer_8,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"HANDLER",7,0,1,0,0,0,0,0,0,0,0,0,0},
      {"IGNDFL",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
   {"pathf90_signal4",15,0,1,0,1,0,0,0,1,0,Signal_Intrinsic,Integer_8,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"HANDLER",7,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"pathf90_signal8",15,0,1,0,1,0,0,0,1,0,Signal_Intrinsic,Integer_8,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"HANDLER",7,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"SIGNAL" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,4,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_signal",14,0,0,0,1,0,0,0,1,0,Signal_Intrinsic,0,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"HANDLER",7,0,1,0,0,0,0,0,0,0,0,0,0},
      {"IGNDFL",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
   {"pathf90_signal4",15,0,0,0,1,0,0,0,1,0,Signal_Intrinsic,0,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"HANDLER",7,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"pathf90_signal8",15,0,0,0,1,0,0,0,1,0,Signal_Intrinsic,0,0},
      {"NUMBER",6,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"HANDLER",7,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
#endif /* KEY Bug 1603 */
{"SIN",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"SIN",3,1,1,0,0,0,0,0,0,0,Sin_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_COMPLEX_MASK,0},
{"SIND",4,1,1,1,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"SIND",4,1,1,0,0,0,0,0,1,0,Sind_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"SINH",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"SINH",4,1,1,0,0,0,0,0,0,0,Sinh_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"SIZE",4,0,1,0,1,0,0,1,0,28,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_4",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SIZE_8",7,0,1,0,1,0,0,0,0,0,Size_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"SIZEOF",6,0,1,0,1,0,0,1,0,18,0,0,TRADITIONAL_FAMILY},
   {"_SIZEOF_4",9,0,1,0,0,0,0,0,1,0,Len_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,0,0,CHARACTER_MASK,0},
   {"_SIZEOF_8",9,0,1,0,0,0,0,0,1,0,Len_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,0,0,CHARACTER_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,0,0,IRCL_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,0,0,IRCL_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,1,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,1,0,IRCLC_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,2,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,2,0,IRCLC_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,3,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,3,0,IRCLC_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,4,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,4,0,IRCLC_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,5,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,5,0,IRCLC_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,6,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,6,0,IRCLC_MASK,0},
   {"_SIZEOF_4",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_4,0},
      {"X",1,0,0,0,0,0,1,0,0,7,0,IRCLC_MASK,0},
   {"_SIZEOF_8",9,0,1,0,1,0,0,0,1,0,Sizeof_Intrinsic,Integer_8,0},
      {"X",1,0,0,0,0,0,1,0,0,7,0,IRCLC_MASK,0},
#ifdef KEY /* Bug 1683 */
{"SLEEP", 5,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_sleep",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"SECONDS",7,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"SNGL",4,1,1,0,0,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"SNGL",4,1,1,0,0,0,0,0,0,0,Sngl_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S8_MASK,0},
   {"SNGL",4,1,1,0,0,0,0,0,0,0,Sngl_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"SNGLQ",5,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SNGLQ",5,1,1,0,0,0,0,0,0,0,Snglq_Intrinsic,Real_4,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,S16_MASK,0},
{"SPACING",7,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"SPACING",7,1,1,0,0,0,0,0,0,0,Spacing_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"SPREAD",6,0,1,0,0,0,0,1,0,7,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,1,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,2,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,3,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,4,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,5,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,6,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_SPREAD",7,0,1,0,1,0,1,0,0,7,Spread_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
      {"NCOPIES",7,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"SQRT",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"SQRT",4,1,1,0,0,0,0,0,0,0,Sqrt_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_COMPLEX_MASK,0},
#ifdef KEY /* Bug 1683 */
{"SRAND", 5,0,0,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"G77_srand_0",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"SEED",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"G77_srand_1",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_8,0},
      {"SEED",4,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
#endif /* KEY Bug 1683 */
#ifdef KEY /* Bug 1683 */
{"STAT",4,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_stat",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"FILE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"STAT" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_stat",12,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"FILE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SARRAY",6,0,0,0,0,0,0,0,0,1,0,I4_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"SUB_AND_FETCH",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"SUB_AND_FETCH",13,1,1,0,0,0,0,0,1,0,Sub_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"SUB_AND_FETCH",13,1,1,0,0,0,0,0,1,0,Sub_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
{"SUM",3,0,1,0,0,0,0,1,0,420,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I1",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I1",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_1,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I1_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I2",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I2",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_2,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I2_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__I4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__I4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__J",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__J",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,I8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__S",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__S",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__D",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__D",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Real_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,S16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C4",9,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C4",8,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C4_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM0__C",8,0,1,0,1,0,0,0,0,0,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__C",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C8_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,0,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,1,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,2,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,3,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,4,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,5,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,0,0,LOGICAL_MASK,0},
   {"_SUM__Z",7,0,1,0,1,0,1,0,0,6,Sum_Intrinsic,Complex_16,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,C16_MASK,0},
      {"DIM",3,0,0,0,0,0,1,0,0,0,0,INTEGER_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
#ifdef KEY /* Bug 1683 */
{"SYMLNK",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_symlnk",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"PATH1",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"PATH2",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"SYMLNK" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_symlnk",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"PATH1",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"PATH2",5,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
{"SYNCHRONIZE",11,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"SYNCHRONIZE",11,1,0,0,0,0,0,0,0,0,Synchronize_Intrinsic,0,0},
# endif
{"SYNC_IMAGES",11,0,0,0,0,0,0,1,0,3,0,0,TRADITIONAL_FAMILY},
   {"shmem_barrier_all",17,0,0,0,1,0,0,0,1,0,Sync_Images_Intrinsic,0,0},
   {"_PAIR_SYNC",10,0,0,0,1,0,0,0,1,0,Sync_Images_Intrinsic,0,0},
      {"IMAGE",5,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_LIST_SYNC",10,0,0,0,1,0,0,0,1,0,Sync_Images_Intrinsic,0,0},
      {"IMAGE",5,0,0,0,0,0,1,0,0,1,0,INTEGER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"SYSTEM",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_system",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"COMMAND",7,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"SYSTEM" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_system",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"COMMAND",7,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"SYSTEM_CLOCK",12,0,0,0,1,0,0,1,0,2,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
#ifdef KEY /* Bug 6009 */
   {"_SYSTEM_CLOCK_4",15,0,0,0,1,0,0,0,0,0,System_Clock_Intrinsic,0,0},
      {"COUNT",5,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
      {"COUNT_RATE",10,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
      {"COUNT_MAX",9,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
   {"_SYSTEM_CLOCK_8",15,0,0,0,1,0,0,0,0,0,System_Clock_Intrinsic,0,0},
      {"COUNT",5,0,0,0,0,1,0,0,0,0,0,I8_MASK,0},
      {"COUNT_RATE",10,0,0,0,0,1,0,0,0,0,0,I8_MASK,0},
      {"COUNT_MAX",9,0,0,0,0,1,0,0,0,0,0,I8_MASK,0},
#else /* KEY Bug 6009 */
   {"_SYSTEM_CLOCK_4",15,0,0,0,1,0,0,0,0,0,0,Integer_4,0},
      {"COUNT",5,0,0,0,0,1,0,0,0,0,0,(I1_MASK|I2_MASK|I4_MASK),0},
      {"COUNT_RATE",10,0,0,0,0,1,0,0,0,0,0,(I1_MASK|I2_MASK|I4_MASK),0},
      {"COUNT_MAX",9,0,0,0,0,1,0,0,0,0,0,(I1_MASK|I2_MASK|I4_MASK),0},
   {"_SYSTEM_CLOCK_8",15,0,0,0,1,0,0,0,0,0,0,Integer_8,0},
      {"COUNT",5,0,0,0,0,1,0,0,0,0,0,I8_MASK,0},
      {"COUNT_RATE",10,0,0,0,0,1,0,0,0,0,0,I8_MASK,0},
      {"COUNT_MAX",9,0,0,0,0,1,0,0,0,0,0,I8_MASK,0},
#endif /* KEY Bug 6009 */
{"TAN",3,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"TAN",3,1,1,0,0,0,0,0,0,0,Tan_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"TAND",4,1,1,0,0,0,1,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY},
   {"TAND",4,1,1,0,0,0,0,0,0,0,Tand_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"TANH",4,1,1,1,0,0,1,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"TANH",4,1,1,0,0,0,0,0,0,0,Tanh_Intrinsic,Real_4,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"TEST_IEEE_EXCEPTION",19,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"TEST_IEEE_EXCEPTION",19,1,1,0,0,0,0,0,0,0,Test_Ieee_Exception_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"EXCEPTION",9,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"EXCEPTION",9,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"TEST_IEEE_INTERRUPT",19,1,1,0,0,1,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"TEST_IEEE_INTERRUPT",19,1,1,0,0,0,0,0,0,0,Test_Ieee_Interrupt_Intrinsic,0,0},
# if defined(_TARGET32_OR_64)
      {"INTERRUPT",9,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
# else
      {"INTERRUPT",9,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
# endif
{"THIS_IMAGE",10,0,1,0,1,0,0,1,0,17,0,0,TRADITIONAL_FAMILY},
   {"_THIS_IMAGE0",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,0,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,1,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,2,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,3,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,4,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,5,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,6,0,                   ALL_MASK,0},
   {"_THIS_IMAGE2",12,0,1,0,1,0,1,0,1,1,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,7,0,                   ALL_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,0,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,1,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,2,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,3,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,4,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,5,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,6,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
   {"_THIS_IMAGE3",12,0,1,0,1,0,0,0,1,0,This_Image_Intrinsic,0,0},
      {"ARRAY",    5,0,0,0,0,0,1,0,0,7,0,                   ALL_MASK,0},
      {"DIM"  ,    3,0,0,0,0,0,0,0,0,0,0,                   INTEGER_MASK,0},
#ifdef KEY /* Bug 4135 */
{"TIME",4,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"pathf90_time4",13,0,1,0,1,0,0,0,1,0,Time4_Intrinsic,Integer_4,0},
{"TIME8",5,0,1,0,1,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"pathf90_time8",13,0,1,0,1,0,0,0,1,0,Time8_Intrinsic,Integer_8,0},
{"TIME" INTRIN_SUBR_SUFFIX,(4 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_subr_time",17,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#else
{"TIME",4,0,0,0,1,0,0,1,0,8,0,0,0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,0,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,1,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,2,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,3,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,4,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,5,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,6,0,(IRC_MASK|CHARACTER_MASK),0},
   {"_TIME",5,0,0,0,1,0,0,0,1,0,Time_Intrinsic,0,0},
      {"BUF",1,0,0,0,0,0,1,0,0,7,0,(IRC_MASK|CHARACTER_MASK),0},
#endif /* KEY Bug 4135 */
#ifdef KEY /* Bug 12813 */
{"TIMEF",5,0,1,0,1,0,0,1,0,1,0,0,0 /* No family */},
   {"_Timef",6,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Real_8,0},
#endif /* KEY Bug 12813 */
{"TINY",4,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"TINY",4,1,1,0,0,0,0,0,0,0,Tiny_Intrinsic,0,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,REAL_MASK,0},
{"TRANSFER",8,0,1,0,1,0,0,1,0,128,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,0,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_TRANSFER",9,0,1,0,1,0,1,0,0,1,Transfer_Intrinsic,0,0},
      {"SOURCE",6,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"MOLD",4,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"SIZE",4,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
{"TRANSPOSE",9,0,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_TRANS",6,0,1,0,1,0,1,0,0,2,Transpose_Intrinsic,0,0},
      {"MATRIX",6,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
{"TRIM",4,0,1,0,1,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))
   {"_F90_TRIM",9,0,1,0,1,0,1,0,0,0,Trim_Intrinsic,0,0},
# else
   {"_TRIM",5,0,1,0,1,0,1,0,0,0,Trim_Intrinsic,0,0},
# endif
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#ifdef KEY /* Bug 1683 */
{"TTYNAM",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"__ttynam_f90",12,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Character_1,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"TTYNAM" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_ttynam",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"UNIT",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"NAME",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
#endif /* KEY Bug 1683 */
{"UBOUND",6,0,1,0,1,0,0,1,0,21,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_4",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND0_8",10,0,1,0,1,0,0,0,0,0,Ubound_Intrinsic,Integer_8,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
      {"DIM",3,0,0,0,0,0,0,0,0,0,0,INTEGER_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_UBOUND",7,0,1,0,1,0,1,0,0,1,Ubound_Intrinsic,Integer_4,0},
      {"ARRAY",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
#ifdef KEY /* Bug 1683 */
{"UMASK",5,0,1,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"G77_umask_0",11,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"MASK",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
{"UMASK" INTRIN_SUBR_SUFFIX,(5 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_umask",13,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"MASK",4,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"OLD",3,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"UNIT",4,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"UNIT",4,1,1,0,0,0,0,0,1,0,Unit_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,INTEGER_TYPELESS_MASK,0},
#ifdef KEY /* Bug 1683 */
{"UNLINK",6,0,1,0,1,0,0,1,0,1,0,0,PGI_FAMILY|G77_FAMILY},
   {"pathf90_unlink",14,0,1,0,1,0,0,0,1,0,Pathf90_Intrinsic,Integer_4,0},
      {"FILE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
{"UNLINK" INTRIN_SUBR_SUFFIX,(6 + (sizeof INTRIN_SUBR_SUFFIX) - 1),0,0,0,1,0,0,1,0,1,0,0,G77_FAMILY},
   {"pathf90_unlink",14,0,0,0,1,0,0,0,1,0,Pathf90_Intrinsic,0,0},
      {"FILE",4,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"STATUS",6,0,0,0,0,1,0,0,0,0,0,I4_MASK,0},
#endif /* KEY Bug 1683 */
{"UNPACK",6,0,1,0,1,0,0,1,0,14,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,1,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,2,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,3,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,4,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,5,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,6,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,7,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,0,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,1,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,1,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,2,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,2,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,2,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,3,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,3,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,3,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,4,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,4,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,4,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,5,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,5,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,5,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,6,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,6,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,6,0,ALL_MASK,0},
   {"_UNPACK",7,0,1,0,1,0,1,0,0,7,Unpack_Intrinsic,0,0},
      {"VECTOR",6,0,0,0,0,0,1,0,0,1,0,ALL_MASK,0},
      {"MASK",4,0,0,0,0,0,1,0,0,7,0,LOGICAL_MASK,0},
      {"FIELD",5,0,0,0,0,0,1,0,0,7,0,ALL_MASK,0},
{"VERIFY",6,1,1,0,0,0,0,1,0,1,0,0,ANSI_FAMILY|TRADITIONAL_FAMILY|PGI_FAMILY},
#ifdef KEY /* Bug 2611 */
   /* Imitate behavior of "index" with respect to result type */
   {"VERIFY",6,1,1,0,0,0,0,0,0,0,Verify_Intrinsic,Integer_4,0},
#else /* KEY Bug 2611 */
   {"VERIFY",6,1,1,0,0,0,0,0,0,0,Verify_Intrinsic,0,0},
#endif /* KEY Bug 2611 */
      {"STRING",6,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"SET",3,0,0,0,0,0,0,0,0,0,0,CHARACTER_MASK,0},
      {"BACK",4,0,0,0,0,1,0,0,0,0,0,LOGICAL_MASK,0},
{"WRITE_MEMORY_BARRIER",20,1,0,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY},
   {"WRITE_MEMORY_BARRIER",20,1,0,0,0,0,0,0,1,0,
                               Write_Memory_Barrier_Intrinsic,0,0},
{"XOR",3,1,1,0,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|PGI_FAMILY|G77_FAMILY},
   {"XOR",3,1,1,0,0,0,0,0,1,0,Xor_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,IRP_MASK,0},
{"XOR_AND_FETCH",13,1,1,0,0,0,0,1,0,2,0,0,TRADITIONAL_FAMILY},
   {"XOR_AND_FETCH",13,1,1,0,0,0,0,0,1,0,Xor_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I4_MASK,0},
   {"XOR_AND_FETCH",13,1,1,0,0,0,0,0,1,0,Xor_And_Fetch_Intrinsic,0,0},
      {"I",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
      {"J",1,0,0,0,0,0,0,0,0,0,0,I8_MASK,0},
#ifdef KEY
{"ZABS",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"ZABS",4,1,1,0,0,0,0,0,1,0,Zabs_Intrinsic,Real_8,0},
      {"A",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"ZCOS",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"ZCOS",4,1,1,0,0,0,0,0,1,0,Zcos_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"ZEXP",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"ZEXP",4,1,1,0,0,0,0,0,1,0,Zexp_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"ZLOG",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"ZLOG",4,1,1,0,0,0,0,0,1,0,Zlog_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"ZSIN",4,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"ZSIN",4,1,1,0,0,0,0,0,1,0,Zsin_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0},
{"ZSQRT",5,1,1,1,0,0,0,1,0,1,0,0,TRADITIONAL_FAMILY|G77_FAMILY},
   {"ZSQRT",5,1,1,0,0,0,0,0,1,0,Zsqrt_Intrinsic,Complex_8,0},
      {"X",1,0,0,0,0,0,0,0,0,0,0,C8_MASK,0}
#endif
};
unsigned MAX_INTRIN_TBL_SIZE = ((sizeof intrin_tbl) / (sizeof *intrin_tbl));

void	(*intrinsic_semantics[]) ()	= {
	unknown_intrinsic,          	/*  Unknown_Intrinsic       	     */
	abs_intrinsic,          	/*  Abs_Intrinsic           	     */
	char_intrinsic,          	/*  Achar_Intrinsic           	     */
	sin_intrinsic,	         	/*  Acos_Intrinsic         	     */
	sin_intrinsic,	         	/*  Acosd_Intrinsic         	     */
	fetch_and_add_intrinsic,       	/*  Add_And_Fetch_Intrinsic          */
	adjustl_intrinsic,      	/*  Adjustl_Intrinsic       	     */
	adjustl_intrinsic,      	/*  Adjustr_Intrinsic       	     */
	aimag_intrinsic,        	/*  Aimag_Intrinsic         	     */
	aint_intrinsic,         	/*  Aint_Intrinsic          	     */
	all_intrinsic,          	/*  All_Intrinsic           	     */
	allocated_intrinsic,    	/*  Allocated_Intrinsic     	     */
	sin_intrinsic,         		/*  Alog_Intrinsic           	     */
	sin_intrinsic,         		/*  Alog10_Intrinsic           	     */
	max_intrinsic,         		/*  Amax0_Intrinsic           	     */
	max_intrinsic,         		/*  Amax1_Intrinsic           	     */
	max_intrinsic,         		/*  Amin0_Intrinsic           	     */
	max_intrinsic,         		/*  Amin1_Intrinsic           	     */
	mod_intrinsic,         		/*  Amod_Intrinsic           	     */
	iand_intrinsic,         	/*  And_Intrinsic           	     */
	fetch_and_add_intrinsic,       	/*  And_And_Fetch_Intrinsic          */
	anint_intrinsic,        	/*  Anint_Intrinsic         	     */
	all_intrinsic,			/*  Any_Intrinsic           	     */
	sin_intrinsic,	         	/*  Asin_Intrinsic          	     */
	sin_intrinsic,	         	/*  Asind_Intrinsic          	     */
	associated_intrinsic,   	/*  Associated_Intrinsic    	     */
	sin_intrinsic,	         	/*  Atan_Intrinsic          	     */
	atan2_intrinsic,        	/*  Atan2_Intrinsic          	     */
	atan2_intrinsic,        	/*  Atan2d_Intrinsic          	     */
	sin_intrinsic,	         	/*  Atand_Intrinsic          	     */
	btest_intrinsic,     		/*  Bitest_Intrinsic      	     */
	bit_size_intrinsic,     	/*  Bit_Size_Intrinsic      	     */
	btest_intrinsic,     		/*  Bjtest_Intrinsic      	     */
	btest_intrinsic,     		/*  Bktest_Intrinsic      	     */
	btest_intrinsic,        	/*  Btest_Intrinsic         	     */
	abs_intrinsic,        		/*  Cabs_Intrinsic         	     */
	sin_intrinsic,        		/*  Ccos_Intrinsic         	     */
	abs_intrinsic,        		/*  Cdabs_Intrinsic         	     */
	sin_intrinsic,        		/*  Cdcos_Intrinsic         	     */
	sin_intrinsic,        		/*  Cdexp_Intrinsic         	     */
	sin_intrinsic,        		/*  Cdlog_Intrinsic         	     */
	sin_intrinsic,        		/*  Cdsin_Intrinsic         	     */
	sin_intrinsic,        		/*  Cdsqrt_Intrinsic         	     */
	ceiling_intrinsic,      	/*  Ceiling_Intrinsic       	     */
	sin_intrinsic,      		/*  Cexp_Intrinsic       	     */
	char_intrinsic,         	/*  Char_Intrinsic          	     */
	set_ieee_exception_intrinsic,   /*  Clear_Ieee_Exception_Intrinsic   */
	loc_intrinsic,         		/*  Cloc_Intrinsic          	     */
	clock_intrinsic,		/*  Clock_Intrinsic         	     */
	sin_intrinsic,			/*  Clog_Intrinsic         	     */
	cmplx_intrinsic,        	/*  Cmplx_Intrinsic         	     */
	compare_and_swap_intrinsic,     /*  Compare_And_Swap_Intrinsic 	     */
	not_intrinsic,        	        /*  Compl_Intrinsic         	     */
	conjg_intrinsic,        	/*  Conjg_Intrinsic         	     */
	sin_intrinsic,          	/*  Cos_Intrinsic           	     */
	sin_intrinsic,          	/*  Cosd_Intrinsic           	     */
	sin_intrinsic,	         	/*  Cosh_Intrinsic          	     */
	sin_intrinsic,          	/*  Cot_Intrinsic           	     */
	all_intrinsic,			/*  Count_Intrinsic         	     */
	unknown_intrinsic,		/*  Cputime_Intrinsic         	     */
	abs_intrinsic,			/*  Cqabs_Intrinsic         	     */
	sin_intrinsic,			/*  Cqcos_Intrinsic         	     */
	sin_intrinsic,			/*  Cqexp_Intrinsic         	     */
	sin_intrinsic,			/*  Cqlog_Intrinsic         	     */
	sin_intrinsic,			/*  Cqsin_Intrinsic         	     */
	sin_intrinsic,			/*  Cqsqrt_Intrinsic         	     */
	cshift_intrinsic,       	/*  Cshift_Intrinsic        	     */
	sin_intrinsic,         		/*  Csin_Intrinsic          	     */
	csmg_intrinsic,         	/*  Csmg_Intrinsic          	     */
	sin_intrinsic,         		/*  Csqrt_Intrinsic          	     */
	cvmgp_intrinsic,        	/*  Cvmgm_Intrinsic         	     */
	cvmgp_intrinsic,        	/*  Cvmgn_Intrinsic         	     */
	cvmgp_intrinsic,        	/*  Cvmgp_Intrinsic         	     */
	cvmgp_intrinsic,        	/*  Cvmgt_Intrinsic         	     */
	cvmgp_intrinsic,        	/*  Cvmgz_Intrinsic         	     */
#ifdef KEY /* Bug 14150 */
        c_f_pointer_intrinsic,		/*  C_F_Pointer_Intrinsic	     */
        c_f_pointer_intrinsic,		/*  C_F_Procpointer_Intrinsic	     */
        loc_intrinsic,			/*  C_Funloc_Intrinsic		     */
        loc_intrinsic,			/*  C_Loc_Iso_Intrinsic	             */
#endif /* KEY Bug 14150 */
	loc_intrinsic,         		/*  C_Loc_Intrinsic          	     */
	abs_intrinsic,			/*  Dabs_Intrinsic          	     */
	sin_intrinsic,			/*  Dacos_Intrinsic          	     */
	sin_intrinsic,			/*  Dacosd_Intrinsic          	     */
	sin_intrinsic,			/*  Dasin_Intrinsic          	     */
	sin_intrinsic,			/*  Dasind_Intrinsic          	     */
	sin_intrinsic,			/*  Datan_Intrinsic          	     */
	atan2_intrinsic,		/*  Datan2_Intrinsic          	     */
	atan2_intrinsic,		/*  Datan2d_Intrinsic          	     */
	sin_intrinsic,			/*  Datand_Intrinsic          	     */
	clock_intrinsic,		/*  Date_Intrinsic          	     */
	unknown_intrinsic,		/*  Date_And_Time_Intrinsic          */
	real_intrinsic,         	/*  Dble_Intrinsic          	     */
	real_intrinsic,         	/*  Dbleq_Intrinsic          	     */
	cmplx_intrinsic,         	/*  Dcmplx_Intrinsic          	     */
	conjg_intrinsic,         	/*  Dconjg_Intrinsic          	     */
	sin_intrinsic,         		/*  Dcos_Intrinsic          	     */
	sin_intrinsic,         		/*  Dcosd_Intrinsic          	     */
	sin_intrinsic,         		/*  Dcosh_Intrinsic          	     */
	sin_intrinsic,         		/*  Dcot_Intrinsic          	     */
	dim_intrinsic,         		/*  Ddim_Intrinsic          	     */
#ifdef KEY /* Bug 1329 */
	erf_intrinsic,			/*  Derf_Intrinsic		     */
	erf_intrinsic,			/*  Derfc_Intrinsic		     */
#endif /* KEY Bug 1329 */
	sin_intrinsic,         		/*  Dexp_Intrinsic          	     */
	real_intrinsic,         	/*  Dfloat_Intrinsic          	     */
	real_intrinsic,         	/*  Dfloati_Intrinsic          	     */
	real_intrinsic,         	/*  Dfloatj_Intrinsic          	     */
	real_intrinsic,         	/*  Dfloatk_Intrinsic          	     */
	digits_intrinsic,       	/*  Digits_Intrinsic        	     */
	dim_intrinsic,          	/*  Dim_Intrinsic           	     */
	aimag_intrinsic,          	/*  Dimag_Intrinsic           	     */
	aint_intrinsic,          	/*  Dint_Intrinsic           	     */
	set_ieee_exception_intrinsic,   /*  Disable_Ieee_Interrupt_Intrinsic */
	sin_intrinsic,          	/*  Dlog_Intrinsic           	     */
	sin_intrinsic,          	/*  Dlog10_Intrinsic           	     */
	max_intrinsic,  		/*  Dmax1_Intrinsic   	     	     */
	max_intrinsic,  		/*  Dmin1_Intrinsic   	     	     */
	mod_intrinsic,  		/*  Dmod_Intrinsic   	     	     */
	anint_intrinsic,  		/*  Dnint_Intrinsic   	     	     */
	dot_product_intrinsic,  	/*  Dot_Product_Intrinsic   	     */
	dprod_intrinsic,        	/*  Dprod_Intrinsic         	     */
	real_intrinsic,        		/*  Dreal_Intrinsic         	     */
	dshiftl_intrinsic,      	/*  Dshiftl_Intrinsic       	     */
	dshiftl_intrinsic,      	/*  Dshiftr_Intrinsic       	     */
	sign_intrinsic,      		/*  Dsign_Intrinsic       	     */
	sin_intrinsic,      		/*  Dsin_Intrinsic       	     */
	sin_intrinsic,      		/*  Dsind_Intrinsic       	     */
	sin_intrinsic,      		/*  Dsinh_Intrinsic       	     */
        dsm_numthreads_intrinsic,       /*  DSM_Chunksize_Intrinsic          */
        dsm_numthreads_intrinsic,       /*  DSM_Distribution_Block_Intrinsic */
        dsm_numthreads_intrinsic,       /*  DSM_Distribution_Cyclic_Intrinsic*/
        dsm_numthreads_intrinsic,       /*  DSM_Distribution_Star_Intrinsic  */
        dsm_numthreads_intrinsic,       /*  DSM_Isdistributed_Intrinsic      */
        dsm_numthreads_intrinsic,       /*  DSM_Isreshaped_Intrinsic         */
        dsm_numthreads_intrinsic,       /*  DSM_Numchunks_Intrinsic          */
        dsm_numthreads_intrinsic,       /*  DSM_Numthreads_Intrinsic         */
        dsm_numthreads_intrinsic,       /*  DSM_Rem_Chunksize_Intrinsic      */
        dsm_numthreads_intrinsic,       /*  DSM_This_Chunksize_Intrinsic     */
        dsm_numthreads_intrinsic,       /*  DSM_This_Startingindex_Intrinsic */
        dsm_numthreads_intrinsic,       /*  DSM_This_Threadnum_Intrinsic     */
        sin_intrinsic,       		/*  Dsqrt_Intrinsic     	     */
        sin_intrinsic,       		/*  Dtan_Intrinsic     	     	     */
        sin_intrinsic,       		/*  Dtand_Intrinsic     	     */
        sin_intrinsic,       		/*  Dtanh_Intrinsic     	     */
# ifdef KEY
        dtime_intrinsic,                /*  Dtime_Intrinsic                  */
# endif
	set_ieee_exception_intrinsic,   /*  Enable_Ieee_Interrupt_Intrinsic  */
	eoshift_intrinsic,      	/*  Eoshift_Intrinsic       	     */
	epsilon_intrinsic,      	/*  Epsilon_Intrinsic       	     */
	iand_intrinsic,          	/*  Eqv_Intrinsic           	     */
#ifdef KEY /* Bug 1324 */
	erf_intrinsic,			/*  Erf_Intrinsic		     */
	erf_intrinsic,			/*  Erfc_Intrinsic		     */
#endif /* KEY Bug 1324 */
#ifdef KEY /* Bug 3018 */
	dtime_intrinsic,		/*  Etime_Intrinsic		     */
#endif /* KEY Bug 3018 */
	exit_intrinsic,          	/*  Exit_Intrinsic           	     */
	sin_intrinsic,          	/*  Exp_Intrinsic           	     */
	exponent_intrinsic,     	/*  Exponent_Intrinsic      	     */
	fcd_intrinsic,          	/*  Fcd_Intrinsic           	     */
# ifdef KEY
	clock_intrinsic,		/*  Fdate_Intrinsic          	     */
# endif
	fetch_and_add_intrinsic,  	/*  Fetch_And_Add_Intrinsic  	     */
	fetch_and_add_intrinsic,  	/*  Fetch_And_And_Intrinsic  	     */
	fetch_and_add_intrinsic,  	/*  Fetch_And_Nand_Intrinsic  	     */
	fetch_and_add_intrinsic,  	/*  Fetch_And_Or_Intrinsic  	     */
	fetch_and_add_intrinsic,  	/*  Fetch_And_Sub_Intrinsic  	     */
	fetch_and_add_intrinsic,  	/*  Fetch_And_Xor_Intrinsic  	     */
	real_intrinsic,        		/*  Float_Intrinsic         	     */
	real_intrinsic,        		/*  Floati_Intrinsic         	     */
	real_intrinsic,        		/*  Floatj_Intrinsic         	     */
	real_intrinsic,        		/*  Floatk_Intrinsic         	     */
	floor_intrinsic,        	/*  Floor_Intrinsic         	     */
#ifdef KEY
	fnum_intrinsic,        	        /*  Fnum_Intrinsic         	     */
#endif
	ieee_finite_intrinsic,        	/*  Fp_Class_Intrinsic       	     */
	fraction_intrinsic,     	/*  Fraction_Intrinsic      	     */
	free_intrinsic, 	    	/*  Free_Intrinsic      	     */
#ifdef KEY
	fstat_intrinsic,	        /*  Fstat_Intrinsic          	     */
#endif
	getpos_intrinsic,       	/*  Getpos_Intrinsic        	     */
	get_ieee_status_intrinsic,      /*  Get_Ieee_Exceptions_Intrinsic    */
	get_ieee_status_intrinsic,      /*  Get_Ieee_Interrupts_Intrinsic    */
	get_ieee_status_intrinsic,      /*  Get_Ieee_Rounding_Mode_Intrinsic */
	get_ieee_status_intrinsic,      /*  Get_Ieee_Status_Intrinsic        */
	huge_intrinsic,         	/*  Huge_Intrinsic          	     */
	abs_intrinsic,         		/*  Iabs_Intrinsic          	     */
	ichar_intrinsic,         	/*  Iachar_Intrinsic          	     */
	iand_intrinsic,         	/*  Iand_Intrinsic          	     */
	ibset_intrinsic,        	/*  Ibchng_Intrinsic         	     */
	ibset_intrinsic,        	/*  Ibclr_Intrinsic         	     */
	ibits_intrinsic,        	/*  Ibits_Intrinsic         	     */
	ibset_intrinsic,        	/*  Ibset_Intrinsic         	     */
	ichar_intrinsic,        	/*  Ichar_Intrinsic         	     */
	idate_intrinsic,        	/*  Idate_Intrinsic         	     */
	dim_intrinsic,        		/*  Idim_Intrinsic         	     */
	int_intrinsic,        		/*  Idint_Intrinsic         	     */
	nint_intrinsic,        		/*  Idnint_Intrinsic         	     */
	ieee_real_intrinsic,  	   	/*  Ieee_Binary_Scale_Intrinsic      */
	ieee_finite_intrinsic,   	/*  Ieee_Class_Intrinsic   	     */
	ieee_real_intrinsic,        	/*  Ieee_Copy_Sign_Intrinsic         */
	ieee_real_intrinsic,        	/*  Ieee_Exponent_Intrinsic          */
	ieee_finite_intrinsic,   	/*  Ieee_Finite_Intrinsic  	     */
	ieee_real_intrinsic,		/*  Ieee_Int_Intrinsic  	     */
	ieee_finite_intrinsic,   	/*  Ieee_Is_Nan_Intrinsic 	     */
	ieee_real_intrinsic,      	/*  Ieee_Next_After_Intrinsic  	     */
	ieee_real_intrinsic,      	/*  Ieee_Real_Intrinsic  	     */
	ieee_real_intrinsic,      	/*  Ieee_Remainder_Intrinsic   	     */
	ieee_real_intrinsic,      	/*  Ieee_Unordered_Intrinsic         */
	iand_intrinsic,         	/*  Ieor_Intrinsic          	     */
	int_intrinsic,         		/*  Ifix_Intrinsic          	     */
	abs_intrinsic,         		/*  Iiabs_Intrinsic          	     */
	iand_intrinsic,         	/*  Iiand_Intrinsic          	     */
	ibset_intrinsic,         	/*  Iibchng_Intrinsic          	     */
	ibset_intrinsic,         	/*  Iibclr_Intrinsic          	     */
	ibits_intrinsic,         	/*  Iibits_Intrinsic          	     */
	ibset_intrinsic,         	/*  Iibset_Intrinsic          	     */
	dim_intrinsic,         		/*  Iidim_Intrinsic          	     */
	int_intrinsic,        		/*  Iidint_Intrinsic         	     */
	iand_intrinsic,        		/*  Iieor_Intrinsic         	     */
	int_intrinsic,        		/*  Iifix_Intrinsic         	     */
	int_intrinsic,        		/*  Iint_Intrinsic         	     */
	iand_intrinsic,        		/*  Iior_Intrinsic         	     */
	int_intrinsic,        		/*  Iiqint_Intrinsic         	     */
	ishft_intrinsic,        	/*  Iisha_Intrinsic         	     */
	ishftc_intrinsic,        	/*  Iishc_Intrinsic         	     */
	ishft_intrinsic,        	/*  Iishft_Intrinsic         	     */
	ishftc_intrinsic,        	/*  Iishftc_Intrinsic         	     */
	ishft_intrinsic,        	/*  Iishl_Intrinsic         	     */
	sign_intrinsic,        		/*  Iisign_Intrinsic         	     */
	ilen_intrinsic,        		/*  Ilen_Intrinsic         	     */
	aimag_intrinsic,        	/*  Imag_Intrinsic         	     */
	mod_intrinsic,          	/*  Imod_Intrinsic           	     */
	mvbits_intrinsic,        	/*  Imvbits_Intrinsic         	     */
	index_intrinsic,        	/*  Index_Intrinsic         	     */
	nint_intrinsic,        		/*  Inint_Intrinsic         	     */
	not_intrinsic,        		/*  Inot_Intrinsic         	     */
	int_intrinsic,          	/*  Int_Intrinsic           	     */
	int_intrinsic,          	/*  Int1_Intrinsic           	     */
	int_intrinsic,          	/*  Int2_Intrinsic           	     */
	int_intrinsic,          	/*  Int4_Intrinsic           	     */
	int_intrinsic,          	/*  Int8_Intrinsic           	     */
	ieee_real_intrinsic,		/*  Int_Mult_Upper_Intrinsic         */
	iand_intrinsic,          	/*  Ior_Intrinsic           	     */
	int_intrinsic,			/*  Iqint_Intrinsic          	     */
	nint_intrinsic,			/*  Iqnint_Intrinsic          	     */
	rtc_intrinsic,			/*  Irtc_Intrinsic          	     */
	ishft_intrinsic,		/*  Isha_Intrinsic          	     */
	ishftc_intrinsic,		/*  Ishc_Intrinsic          	     */
	ishft_intrinsic,        	/*  Ishft_Intrinsic         	     */
	ishftc_intrinsic,       	/*  Ishftc_Intrinsic        	     */
	ishft_intrinsic,		/*  Ishl_Intrinsic          	     */
	sign_intrinsic,			/*  Isign_Intrinsic          	     */
	ieee_finite_intrinsic,	        /*  Isnan_Intrinsic          	     */
	clock_intrinsic,		/*  Jdate_Intrinsic	    	     */
	abs_intrinsic,			/*  Jiabs_Intrinsic	    	     */
	iand_intrinsic,			/*  Jiand_Intrinsic	    	     */
	ibset_intrinsic,         	/*  Jibchng_Intrinsic          	     */
	ibset_intrinsic,		/*  Jibclr_Intrinsic	    	     */
	ibits_intrinsic,		/*  Jibits_Intrinsic	    	     */
	ibset_intrinsic,		/*  Jibset_Intrinsic	    	     */
	dim_intrinsic,			/*  Jidim_Intrinsic	    	     */
	int_intrinsic,        		/*  Jidint_Intrinsic         	     */
	iand_intrinsic,        		/*  Jieor_Intrinsic         	     */
	int_intrinsic,        		/*  Jifix_Intrinsic         	     */
	int_intrinsic,        		/*  Jint_Intrinsic         	     */
	iand_intrinsic,        		/*  Jior_Intrinsic         	     */
	int_intrinsic,        		/*  Jiqint_Intrinsic         	     */
	ishft_intrinsic,        	/*  Jisha_Intrinsic         	     */
	ishftc_intrinsic,        	/*  Jishc_Intrinsic         	     */
	ishft_intrinsic,        	/*  Jishft_Intrinsic         	     */
	ishftc_intrinsic,        	/*  Jishftc_Intrinsic         	     */
	ishft_intrinsic,        	/*  Jishl_Intrinsic         	     */
	sign_intrinsic,        		/*  Jisign_Intrinsic         	     */
	mod_intrinsic,          	/*  Jmod_Intrinsic           	     */
	mvbits_intrinsic,        	/*  Jmvbits_Intrinsic         	     */
	nint_intrinsic,        		/*  Jnint_Intrinsic         	     */
	not_intrinsic,        		/*  Jnot_Intrinsic         	     */
	abs_intrinsic,        		/*  Kiabs_Intrinsic         	     */
	iand_intrinsic,        		/*  Kiand_Intrinsic         	     */
	ibset_intrinsic,         	/*  Kibchng_Intrinsic          	     */
	ibset_intrinsic,        	/*  Kibclr_Intrinsic         	     */
	ibits_intrinsic,        	/*  Kibits_Intrinsic         	     */
	ibset_intrinsic,        	/*  Kibset_Intrinsic         	     */
	dim_intrinsic,        		/*  Kidim_Intrinsic         	     */
	int_intrinsic,        		/*  Kidint_Intrinsic         	     */
	iand_intrinsic,        		/*  Kieor_Intrinsic         	     */
	int_intrinsic,        		/*  Kifix_Intrinsic         	     */
#ifdef KEY
	kill_intrinsic,        		/*  Kill_Intrinsic         	     */
#endif
	kind_intrinsic,         	/*  Kind_Intrinsic          	     */
	int_intrinsic,        		/*  Kint_Intrinsic         	     */
	iand_intrinsic,        		/*  Kior_Intrinsic         	     */
	int_intrinsic,        		/*  Kiqint_Intrinsic         	     */
	ishft_intrinsic,        	/*  Kisha_Intrinsic         	     */
	ishftc_intrinsic,        	/*  Kishc_Intrinsic         	     */
	ishft_intrinsic,        	/*  Kishft_Intrinsic         	     */
	ishftc_intrinsic,        	/*  Kishftc_Intrinsic         	     */
	ishft_intrinsic,        	/*  Kishl_Intrinsic         	     */
	sign_intrinsic,        		/*  Kisign_Intrinsic         	     */
	mod_intrinsic,          	/*  Kmod_Intrinsic           	     */
	mvbits_intrinsic,        	/*  Kmvbits_Intrinsic         	     */
	nint_intrinsic,        		/*  Knint_Intrinsic         	     */
	not_intrinsic,        		/*  Knot_Intrinsic         	     */
	lbound_intrinsic,       	/*  Lbound_Intrinsic        	     */
	leadz_intrinsic,        	/*  Leadz_Intrinsic         	     */
	len_intrinsic,          	/*  Len_Intrinsic           	     */
	length_intrinsic,       	/*  Length_Intrinsic        	     */
	len_trim_intrinsic,     	/*  Len_Trim_Intrinsic      	     */
	lge_intrinsic,          	/*  Lge_Intrinsic           	     */
	lge_intrinsic,          	/*  Lgt_Intrinsic           	     */
	lge_intrinsic,          	/*  Lle_Intrinsic           	     */
	lge_intrinsic,          	/*  Llt_Intrinsic           	     */
	loc_intrinsic,          	/*  Loc_Intrinsic           	     */
	lock_release_intrinsic,      	/*  Lock_Release_Intrinsic           */
	fetch_and_add_intrinsic,      	/*  Lock_Test_And_Set_Intrinsic      */
	sin_intrinsic,          	/*  Log_Intrinsic           	     */
	sin_intrinsic,	        	/*  Log10_Intrinsic         	     */
        num_images_intrinsic,           /*  Log2_Images_Intrinsic            */
	logical_intrinsic,      	/*  Logical_Intrinsic       	     */
	int_intrinsic,      		/*  Long_Intrinsic       	     */
	shift_intrinsic,      		/*  Lshift_Intrinsic       	     */
	mclr_intrinsic,      		/*  M@clr_Intrinsic       	     */
	mld_intrinsic,      		/*  M@ld_Intrinsic       	     */
	mldmx_intrinsic,      		/*  M@ldmx_Intrinsic       	     */
	mmx_intrinsic,      		/*  M@mx_Intrinsic       	     */
	mul_intrinsic,      		/*  M@ul_Intrinsic       	     */
	malloc_intrinsic,         	/*  Malloc_Intrinsic          	     */
	mask_intrinsic,         	/*  Mask_Intrinsic          	     */
	matmul_intrinsic,       	/*  Matmul_Intrinsic        	     */
	max_intrinsic,          	/*  Max_Intrinsic           	     */
	max_intrinsic,          	/*  Max0_Intrinsic           	     */
	max_intrinsic,          	/*  Max1_Intrinsic           	     */
	maxexponent_intrinsic,  	/*  Maxexponent_Intrinsic   	     */
	minval_intrinsic,		/*  Maxloc_Intrinsic         	     */
	minval_intrinsic,		/*  Maxval_Intrinsic        	     */
	memory_barrier_intrinsic,	/*  Memory_Barrier_Intrinsic	     */
	mergee_intrinsic,        	/*  Merge_Intrinsic         	     */
	max_intrinsic,          	/*  Min_Intrinsic           	     */
	max_intrinsic,          	/*  Min0_Intrinsic           	     */
	max_intrinsic,          	/*  Min1_Intrinsic           	     */
	minexponent_intrinsic,  	/*  Minexponent_Intrinsic   	     */
	minval_intrinsic,       	/*  Minloc_Intrinsic        	     */
	minval_intrinsic,       	/*  Minval_Intrinsic        	     */
	mod_intrinsic,          	/*  Mod_Intrinsic           	     */
	modulo_intrinsic,       	/*  Modulo_Intrinsic       	     */
	mvbits_intrinsic,       	/*  Mvbits_Intrinsic        	     */
        my_pe_intrinsic,                /*  My_Pe_Intrinsic         	     */
	fetch_and_add_intrinsic,       	/*  Nand_And_Fetch_Intrinsic         */
	nearest_intrinsic,      	/*  Nearest_Intrinsic       	     */
	iand_intrinsic,         	/*  Neqv_Intrinsic          	     */
	nint_intrinsic,         	/*  Nint_Intrinsic          	     */
	not_intrinsic,          	/*  Not_Intrinsic           	     */
	null_intrinsic,          	/*  Null_Intrinsic           	     */
	numarg_intrinsic,       	/*  Numarg_Intrinsic        	     */
        num_images_intrinsic,           /*  Num_Images_Intrinsic             */
#ifdef KEY
        unknown_intrinsic,              /*  Omp_Destroy_Lock_Intrinsic       */
        unknown_intrinsic,              /*  Omp_Destroy_Nest_Lock_Intrinsic  */
#endif
	omp_get_max_threads_intrinsic,	/*  Omp_Get_Dynamic_Intrinsic  	     */
	omp_get_max_threads_intrinsic,  /*  Omp_Get_Max_Threads_Intrinsic    */
	omp_get_max_threads_intrinsic,  /*  Omp_Get_Nested_Intrinsic         */
	omp_get_max_threads_intrinsic,  /*  Omp_Get_Num_Procs_Intrinsic      */
	omp_get_max_threads_intrinsic, 	/*  Omp_Get_Num_Threads_Intrinsic    */
	omp_get_max_threads_intrinsic,  /*  Omp_Get_Thread_Num_Intrinsic     */
#ifdef KEY
	omp_get_max_threads_intrinsic,  /*  Omp_Get_Wtick_Intrinsic          */
	omp_get_max_threads_intrinsic,  /*  Omp_Get_Wtime_Intrinsic          */
        unknown_intrinsic,              /*  Omp_Init_Lock_Intrinsic          */
        unknown_intrinsic,              /*  Omp_Init_Nest_Lock_Intrinsic     */
#endif
	omp_get_max_threads_intrinsic,	/*  Omp_In_Parallel_Intrinsic        */
#ifdef KEY
	unknown_intrinsic,       	/*  Omp_Set_Lock_Intrinsic    	     */
	unknown_intrinsic,       	/*  Omp_Set_Nest_Lock_Intrinsic      */
	omp_get_max_threads_intrinsic, 	/*  Omp_Test_Lock_Intrinsic    	     */
	omp_get_max_threads_intrinsic, 	/*  Omp_Test_Nest_Lock_Intrinsic     */
	unknown_intrinsic,       	/*  Omp_Unset_Lock_Intrinsic  	     */
	unknown_intrinsic,       	/*  Omp_Unset_Nest_Lock_Intrinsic    */
#else
	omp_set_lock_intrinsic,       	/*  Omp_Set_Lock_Intrinsic    	     */
	omp_set_lock_intrinsic,       	/*  Omp_Test_Lock_Intrinsic    	     */
	omp_set_lock_intrinsic,       	/*  Omp_Unset_Lock_Intrinsic  	     */
#endif
	iand_intrinsic,         	/*  Or_Intrinsic            	     */
	fetch_and_add_intrinsic,       	/*  Or_And_Fetch_Intrinsic           */
	pack_intrinsic,         	/*  Pack_Intrinsic          	     */
	leadz_intrinsic,       		/*  Popcnt_Intrinsic        	     */
	leadz_intrinsic,       		/*  Poppar_Intrinsic        	     */
	precision_intrinsic,    	/*  Precision_Intrinsic     	     */
	present_intrinsic,      	/*  Present_Intrinsic       	     */
	minval_intrinsic,		/*  Product_Intrinsic       	     */
	abs_intrinsic,			/*  Qabs_Intrinsic       	     */
	sin_intrinsic,			/*  Qacos_Intrinsic       	     */
	sin_intrinsic,			/*  Qacosd_Intrinsic       	     */
	sin_intrinsic,			/*  Qasin_Intrinsic       	     */
	sin_intrinsic,			/*  Qasind_Intrinsic       	     */
	sin_intrinsic,			/*  Qatan_Intrinsic       	     */
	atan2_intrinsic,		/*  Qatan2_Intrinsic       	     */
	atan2_intrinsic,		/*  Qatan2d_Intrinsic       	     */
	sin_intrinsic,			/*  Qatand_Intrinsic       	     */
	cmplx_intrinsic,		/*  Qcmplx_Intrinsic       	     */
	sin_intrinsic,			/*  Qcos_Intrinsic       	     */
	sin_intrinsic,			/*  Qcosd_Intrinsic       	     */
	sin_intrinsic,			/*  Qcosh_Intrinsic       	     */
	sin_intrinsic,			/*  Qcot_Intrinsic       	     */
	dim_intrinsic,			/*  Qdim_Intrinsic       	     */
	sin_intrinsic,			/*  Qexp_Intrinsic       	     */
	real_intrinsic,			/*  Qext_Intrinsic       	     */
	real_intrinsic,			/*  Qfloat_Intrinsic       	     */
	real_intrinsic,			/*  Qfloati_Intrinsic       	     */
	real_intrinsic,			/*  Qfloatj_Intrinsic       	     */
	real_intrinsic,			/*  Qfloatk_Intrinsic       	     */
	aimag_intrinsic,		/*  Qimag_Intrinsic       	     */
	aint_intrinsic,			/*  Qint_Intrinsic       	     */
	sin_intrinsic,			/*  Qlog_Intrinsic       	     */
	sin_intrinsic,			/*  Qlog10_Intrinsic       	     */
	mod_intrinsic,			/*  Qmod_Intrinsic       	     */
	anint_intrinsic,  		/*  Qnint_Intrinsic   	     	     */
	dprod_intrinsic,  		/*  Qprod_Intrinsic   	     	     */
	real_intrinsic,  		/*  Qreal_Intrinsic   	     	     */
	sign_intrinsic,			/*  Qsign_Intrinsic       	     */
	sin_intrinsic,			/*  Qsin_Intrinsic       	     */
	sin_intrinsic,			/*  Qsind_Intrinsic       	     */
	sin_intrinsic,			/*  Qsinh_Intrinsic       	     */
	sin_intrinsic,			/*  Qsqrt_Intrinsic       	     */
	sin_intrinsic,			/*  Qtan_Intrinsic       	     */
	sin_intrinsic,			/*  Qtand_Intrinsic       	     */
	sin_intrinsic,			/*  Qtanh_Intrinsic       	     */
	radix_intrinsic,        	/*  Radix_Intrinsic         	     */
	ranf_intrinsic,        		/*  Ran_Intrinsic         	     */
	random_number_intrinsic,	/*  Random_Number_Intrinsic 	     */
	random_seed_intrinsic,  	/*  Random_Seed_Intrinsic   	     */
	ranf_intrinsic,        		/*  Randu_Intrinsic         	     */
	ranf_intrinsic,         	/*  Ranf_Intrinsic          	     */
	range_intrinsic,        	/*  Range_Intrinsic         	     */
	ranget_intrinsic,       	/*  Ranget_Intrinsic        	     */
	ranget_intrinsic,       	/*  Ranset_Intrinsic        	     */
	readsm_intrinsic,         	/*  Readsm_Intrinsic        	     */
	real_intrinsic,         	/*  Real_Intrinsic          	     */
	remote_write_barrier_intrinsic, /*  R_W_B_Intrinsic         	     */
        num_images_intrinsic,           /*  Rem_Images_Intrinsic             */
	repeat_intrinsic,       	/*  Repeat_Intrinsic        	     */
	reshape_intrinsic,      	/*  Reshape_Intrinsic       	     */
	rrspacing_intrinsic,    	/*  Rrspacing_Intrinsic     	     */
	shift_intrinsic,          	/*  Rshift_Intrinsic           	     */
	rtc_intrinsic,          	/*  Rtc_Intrinsic           	     */
	scale_intrinsic,        	/*  Scale_Intrinsic         	     */
	index_intrinsic,        	/*  Scan_Intrinsic         	     */
	selected_int_kind_intrinsic,   	/*  SIK_Intrinsic           	     */
	selected_real_kind_intrinsic,  	/*  SRK_Intrinsic           	     */
	set_exponent_intrinsic, 	/*  Set_Exponent_Intrinsic  	     */
	set_ieee_exception_intrinsic,   /*  Set_Ieee_Exception_Intrinsic     */
	get_ieee_status_intrinsic,      /*  Set_Ieee_Exceptions_Intrinsic    */
	get_ieee_status_intrinsic,	/*  Set_Ieee_Interrupts_Intrinsic    */
	get_ieee_status_intrinsic,	/*  Set_Ieee_Rounding_Mode_Intrinsic */
	get_ieee_status_intrinsic,	/*  Set_Ieee_Status_Intrinsic        */
	shape_intrinsic,        	/*  Shape_Intrinsic         	     */
	shift_intrinsic,        	/*  Shift_Intrinsic         	     */
	shift_intrinsic,        	/*  Shifta_Intrinsic         	     */
	shift_intrinsic,        	/*  Shiftl_Intrinsic        	     */
	shift_intrinsic,        	/*  Shiftr_Intrinsic        	     */
	int_intrinsic,        		/*  Short_Intrinsic        	     */
	sign_intrinsic,         	/*  Sign_Intrinsic          	     */
#ifdef KEY
	signal_intrinsic,         	/*  Signal_Intrinsic          	     */
#endif
	sin_intrinsic,          	/*  Sin_Intrinsic           	     */
	sin_intrinsic,          	/*  Sind_Intrinsic           	     */
	sin_intrinsic,	         	/*  Sinh_Intrinsic          	     */
	size_intrinsic,			/*  Size_Intrinsic	    	     */
	sizeof_intrinsic,		/*  Sizeof_Intrinsic	    	     */
	real_intrinsic,	         	/*  Sngl_Intrinsic          	     */
	real_intrinsic,	         	/*  Snglq_Intrinsic          	     */
	spacing_intrinsic,      	/*  Spacing_Intrinsic       	     */
	spread_intrinsic,       	/*  Spread_Intrinsic         	     */
	sin_intrinsic,	         	/*  Sqrt_Intrinsic          	     */
#ifdef KEY
	stat_intrinsic,	         	/*  Stat_Intrinsic          	     */
#endif
	fetch_and_add_intrinsic,       	/*  Sub_And_Fetch_Intrinsic          */
	minval_intrinsic,		/*  Sum_Intrinsic           	     */
	synchronize_intrinsic,		/*  Synchronize_Intrinsic            */
        num_images_intrinsic,           /*  Sync_Images_Intrinsic            */
	system_clock_intrinsic,		/*  System_Clock_Intrinsic           */
	sin_intrinsic,          	/*  Tan_Intrinsic           	     */
	sin_intrinsic,          	/*  Tand_Intrinsic           	     */
	sin_intrinsic,	         	/*  Tanh_Intrinsic          	     */
 	test_ieee_interrupt_intrinsic,  /*  Test_Ieee_Exception_Intrinsic    */
 	test_ieee_interrupt_intrinsic,  /*  Test_Ieee_Interrupt_Intrinsic    */
        num_images_intrinsic,           /*  This_Image_Intrinsic             */
	free_intrinsic,         	/*  Time_Intrinsic          	     */
#ifdef KEY
	time_intrinsic,         	/*  Time4_Intrinsic          	     */
	time_intrinsic,         	/*  Time8_Intrinsic          	     */
#endif
	tiny_intrinsic,         	/*  Tiny_Intrinsic          	     */
	transfer_intrinsic,     	/*  Transfer_Intrinsic      	     */
	transpose_intrinsic,    	/*  Transpose_Intrinsic     	     */
	trim_intrinsic,         	/*  Trim_Intrinsic          	     */
	ubound_intrinsic,		/*  Ubound_Intrinsic        	     */
	unit_intrinsic,         	/*  Unit_Intrinsic          	     */
	unpack_intrinsic,	       	/*  Unpack_Intrinsic                 */
	index_intrinsic,	       	/*  Verify_Intrinsic                 */
        write_memory_barrier_intrinsic, /*  W_M_B_Intrinsic          	     */
	iand_intrinsic,  	       	/*  Xor_Intrinsic            	     */
	fetch_and_add_intrinsic, 	/*  Xor_And_Fetch_Intrinsic          */
#ifdef KEY
	abs_intrinsic,        		/*  Zabs_Intrinsic         	     */
	sin_intrinsic,        		/*  Zcos_Intrinsic         	     */
	sin_intrinsic,        		/*  Zexp_Intrinsic         	     */
	sin_intrinsic,        		/*  Zlog_Intrinsic         	     */
	sin_intrinsic,        		/*  Zsin_Intrinsic         	     */
	sin_intrinsic,        		/*  Zsqrt_Intrinsic         	     */
#endif
#ifdef KEY /* Bug 1683 */
	pathf90_intrinsic,		/*  Pathf90_Intrinsic		     */
#endif /* KEY Bug 1683 */
#ifdef KEY /* Bug 5089 */
	true_intrinsic,			/*  True_Intrinsic		     */
	support_uflow_intrinsic,	/*  Support_Uflow_Intrinsic	     */
#endif /* KEY Bug 5089 */
#ifdef KEY /* F2003 */
	newline_intrinsic		/*  Newline_Intrinsic		     */
#endif /* F2003 */
	};


/**************************************************************************\
|* The global Intrinsic mapping table                                     *|
\**************************************************************************/

intrin_map_type         intrin_map[MAX_INTRIN_MAP_SIZE] =

/*              KIND=4          KIND=8                                    */
{

# if (defined(_TARGET_OS_IRIX) || defined(_TARGET_OS_LINUX) || defined(_TARGET_OS_DARWIN))

   {"ABS",	"r_abs",	"d_abs"},
   {"ACOS",	"r_acos",	"d_acos"},
   {"AIMAG",	"r_imag",	"d_imag"},
   {"AINT",	"r_int",	"d_int"},
   {"ALOG",	"r_log",	"d_log"},
   {"ALOG10",	"r_lg10",	"d_lg10"},
   {"AMOD",	"r_mod",	"d_mod"},
   {"ANINT",	"r_nint",	"d_nint"},
   {"ASIN",	"r_asin",	"d_asin"},
   {"ATAN",	"r_atan",	"d_atan"},
   {"ATAN2",	"r_atn2",	"d_atn2"},
   {"CABS",	"c_abs",	"z_abs_"},
   {"CCOS",	"c_cos",	"z_cos"},
   {"CEXP",	"c_exp",	"z_exp"},
   {"CLOG",	"c_log",	"z_log"},
   {"CONJG",	"r_cnjg",	"d_cnjg"},
   {"COS",	"r_cos",	"d_cos"},
   {"COSH",	"r_cosh",	"d_cosh"},
   {"COT",	"r_cot",	"d_cot"},
   {"CSIN",	"c_sin",	"z_sin"},
   {"CSQRT",	"c_sqrt",	"z_sqrt"},
   {"DABS",	"d_abs",	"__q_abs"},
   {"DACOS",	"d_acos",	"__q_acos"},
   {"DASIN",	"d_asin",	"__q_asin"},
   {"DATAN",	"d_atan",	"__q_atan"},
   {"DATAN2",	"d_atn2",	"__q_atn2"},
   {"DCOS",	"d_cos",	"__q_cos"},
   {"DCOSH",	"d_cosh",	"__q_cosh"},
   {"DCOT",	"d_cot",	"__q_cot"},
   {"DDIM",	"d_dim",	"__q_dim"},
   {"DEXP",	"d_exp",	"__q_exp"},
   {"DIM",	"r_dim",	"d_dim"},
   {"DINT",	"d_int",	"__q_int"},
   {"DLOG",	"d_log",	"__q_log"},
   {"DLOG10",	"d_lg10",	"__q_lg10"},
   {"DMOD",	"d_mod",	"__q_mod"},
   {"DNINT",	"d_nint",	"__q_nint"},
   {"DPROD",	"d_prod",	"__q_prod"},
   {"DSIGN",	"d_sign",	"__q_sign"},
   {"DSIN",	"d_sin",	"__q_sin"},
   {"DSINH",	"d_sinh",	"__q_sinh"},
   {"DSQRT",	"d_sqrt",	"__q_sqrt"},
   {"DTAN",	"d_tan",	"__q_tan"},
   {"DTANH",	"d_tanh",	"__q_tanh"},
#ifdef KEY /* Bug 1324 */
   {"ERF",	"erf_",		"derf_"},
   {"ERFC",	"erfc_",	"derfc_"},
   {"DERF",	"derf_",	"__q_erf_"},
   {"DERFC",	"derfc_",	"__q_erfc_"},
#endif /* KEY Bug 1324 */
   {"EXP",	"r_exp",	"d_exp"},
   {"IABS",	"i_abs",	"l_abs"},
   {"IDIM",	"i_dim",	"l_dim"},
   {"IDNINT",	"i_dnnt",	"l_dnnt"},
   {"INDEX",	"i_indx",	"l_indx"},
   {"ISIGN",	"i_sign",	"l_sign"},
   {"LEN",	"i_len",	"l_len"},
   {"MOD",	"i_mod",	"l_mod"},
#ifdef KEY /* Bug 3869 */
   /* Note that NINT is handled specially in p_utils.c */
#endif /* KEY bug 3869 */
   {"NINT",	"i_nint",	"l_nint"},
   {"SIGN",	"r_sign",	"d_sign"},
   {"SIN",	"r_sin",	"d_sin"},
   {"SINH",	"r_sinh",	"d_sinh"},
   {"SQRT",	"r_sqrt",	"d_sqrt"},
   {"TAN",	"r_tan",	"d_tan"},
   {"TANH",	"r_tanh",	"d_tanh"},
#ifdef KEY
   {"ZABS",	"z_abs_",	"z_abs_"},
   {"ZCOS",	"z_cos",	"z_cos"},
   {"ZEXP",	"z_exp",	"z_exp"},
   {"ZLOG",	"z_log",	"z_log"},
   {"ZSIN",	"z_sin",	"z_sin"},
   {"ZSQRT",	"z_sqrt",	"z_sqrt"},
   {"CDABS",	"z_abs_",	"z_abs_"},
   {"CDCOS",	"z_cos",	"z_cos"},
   {"CDEXP",	"z_exp",	"z_exp"},
   {"CDLOG",	"z_log",	"z_log"},
   {"CDSIN",	"z_sin",	"z_sin"},
   {"CDSQRT",	"z_sqrt",	"z_sqrt"}
#endif

# elif defined(_TARGET_SV2)

   {"ABS",      "__HABS_",       "__ABS_"},
   {"ACOS",     "__HACOS_",      "__ACOS_"},
   {"AIMAG",    "__HIMAG_",      "__AIMAG_"},
   {"AINT",     "__HINT_",       "__AINT_"},
   {"ALOG",     "__HLOG_",       "__ALOG_"},
   {"ALOG10",   "__HLOG10_",     "__ALOG10_"},
   {"AMOD",     "__HMOD_",       "__AMOD_"},
   {"ANINT",    "__HNINT_",      "__ANINT_"},
   {"ASIN",     "__HASIN_",      "__ASIN_"},
   {"ATAN",     "__HATAN_",      "__ATAN_"},
   {"ATAN2",    "__HATAN2_",     "__ATAN2_"},
   {"CABS",     "__CHABS_",      "__CABS_"},
   {"CCOS",     "__CHCOS_",      "__CCOS_"},
   {"CEXP",     "__CHEXP_",      "__CEXP_"},
   {"CLOG",     "__CHLOG_",      "__CLOG_"},
   {"CONJG",    "__HCONJG_",     "__CONJG_"},
   {"COS",      "__HCOS_",       "__COS_"},
   {"COSH",     "__HCOSH_",      "__COSH_"},
   {"COT",      "__HCOT_",       "__COT_"},
   {"CSIN",     "__CHSIN_",      "__CSIN_"},
   {"CSQRT",    "__CHSQRT_",     "__CSQRT_"},
   {"DABS",     "__ABS_",        "__DABS_"},
   {"DACOS",    "__ACOS_",       "__DACOS_"},
   {"DASIN",    "__ASIN_",       "__DASIN_"},
   {"DATAN",    "__ATAN_",       "__DATAN_"},
   {"DATAN2",   "__ATAN2_",      "__DATAN2_"},
   {"DCOS",     "__COS_",        "__DCOS_"},
   {"DCOSH",    "__COSH_",       "__DCOSH_"},
   {"DCOT",     "__COT_",        "__DCOT_"},
   {"DDIM",     "__DIM_",        "__DDIM_"},
   {"DEXP",     "__EXP_",        "__DEXP_"},
   {"DIM",      "__HDIM_",       "__DIM_"},
   {"DINT",     "__AINT_",       "__DINT_"},
   {"DLOG",     "__ALOG_",       "__DLOG_"},
   {"DLOG10",   "__ALOG10_",     "__DLOG10_"},
   {"DMOD",     "__AMOD_",       "__DMOD_"},
   {"DNINT",    "__ANINT_",      "__DNINT_"},
   {"DPROD",    "__PROD_",       "__DPROD_"},
   {"DSIGN",    "__SIGN_",       "__DSIGN_"},
   {"DSIN",     "__SIN_",        "__DSIN_"},
   {"DSINH",    "__SINH_",       "__DSINH_"},
   {"DSQRT",    "__SQRT_",       "__DSQRT_"},
   {"DTAN",     "__TAN_",        "__DTAN_"},
   {"DTANH",    "__TANH_",       "__DTANH_"},
   {"EXP",      "__HEXP_",       "__EXP_"},
   {"IABS",     "__NABS_",       "__IABS_"},
   {"IDIM",     "__NDIM_",       "__IDIM_"},
   {"IDNINT",   "__NANINT_",     "__IDNINT_"},
   {"INDEX",    "__INDEX_4_",    "__INDEX_8_"},
   {"ISIGN",    "__NSIGN_",      "__ISIGN_"},
   {"LEN",      "__LEN_4_",      "__LEN_8_"},
   {"MOD",      "__NMOD_",       "__MOD_"},
   {"NINT",     "__NHNINT_",     "__NINT_"},
   {"SIGN",     "__HSIGN_",      "__SIGN_"},
   {"SIN",      "__HSIN_",       "__SIN_"},
   {"SINH",     "__HSINH_",      "__SINH_"},
   {"SQRT",     "__HSQRT_",      "__SQRT_"},
   {"TAN",      "__HTAN_",       "__TAN_"},
   {"TANH",     "__HTANH_",      "__TANH_"}

# else

   {"ABS",	"_HABS_",	"_ABS_"},
   {"ACOS",	"_HACOS_",	"_ACOS_"},
   {"AIMAG",	"_HIMAG_",	"_AIMAG_"},
   {"AINT",	"_HINT_",	"_AINT_"},
   {"ALOG",	"_HLOG_",	"_ALOG_"},
   {"ALOG10",	"_HLOG10_",	"_ALOG10_"},
   {"AMOD",	"_HMOD_",	"_AMOD_"},
   {"ANINT",	"_HNINT_",	"_ANINT_"},
   {"ASIN",	"_HASIN_",	"_ASIN_"},
   {"ATAN",	"_HATAN_",	"_ATAN_"},
   {"ATAN2",	"_HATAN2_",	"_ATAN2_"},
   {"CABS",	"_CHABS_",	"_CABS_"},
   {"CCOS",	"_CHCOS_",	"_CCOS_"},
   {"CEXP",	"_CHEXP_",	"_CEXP_"},
   {"CLOG",	"_CHLOG_",	"_CLOG_"},
   {"CONJG",	"_HCONJG_",	"_CONJG_"},
   {"COS",	"_HCOS_",	"_COS_"},
   {"COSH",	"_HCOSH_",	"_COSH_"},
   {"COT",	"_HCOT_",	"_COT_"},
   {"CSIN",	"_CHSIN_",	"_CSIN_"},
   {"CSQRT",	"_CHSQRT_",	"_CSQRT_"},
   {"DABS",	"_ABS_",	"_DABS_"},
   {"DACOS",	"_ACOS_",	"_DACOS_"},
   {"DASIN",	"_ASIN_",	"_DASIN_"},
   {"DATAN",	"_ATAN_",	"_DATAN_"},
   {"DATAN2",	"_ATAN2_",	"_DATAN2_"},
   {"DCOS",	"_COS_",	"_DCOS_"},
   {"DCOSH",	"_COSH_",	"_DCOSH_"},
   {"DCOT",	"_COT_",	"_DCOT_"},
   {"DDIM",	"_DIM_",	"_DDIM_"},
   {"DEXP",	"_EXP_",	"_DEXP_"},
   {"DIM",	"_HDIM_",	"_DIM_"},
   {"DINT",	"_AINT_",	"_DINT_"},
   {"DLOG",	"_ALOG_",	"_DLOG_"},
   {"DLOG10",	"_ALOG10_",	"_DLOG10_"},
   {"DMOD",	"_AMOD_",	"_DMOD_"},
   {"DNINT",	"_ANINT_",	"_DNINT_"},
   {"DPROD",	"_PROD_",	"_DPROD_"},
   {"DSIGN",	"_SIGN_",	"_DSIGN_"},
   {"DSIN",	"_SIN_",	"_DSIN_"},
   {"DSINH",	"_SINH_",	"_DSINH_"},
   {"DSQRT",	"_SQRT_",	"_DSQRT_"},
   {"DTAN",	"_TAN_",	"_DTAN_"},
   {"DTANH",	"_TANH_",	"_DTANH_"},
   {"EXP",	"_HEXP_",	"_EXP_"},
   {"IABS",	"_NABS_",	"_IABS_"},
   {"IDIM",	"_NDIM_",	"_IDIM_"},
   {"IDNINT",	"_NANINT_",	"_IDNINT_"},
   {"INDEX",	"_INDEX_4_",	"_INDEX_8_"},
   {"ISIGN",	"_NSIGN_",	"_ISIGN_"},
   {"LEN",	"_LEN_4_",	"_LEN_8_"},
   {"MOD",	"_NMOD_",	"_MOD_"},
   {"NINT",	"_NHNINT_",	"_NINT_"},
   {"SIGN",	"_HSIGN_",	"_SIGN_"},
   {"SIN",	"_HSIN_",	"_SIN_"},
   {"SINH",	"_HSINH_",	"_SINH_"},
   {"SQRT",	"_HSQRT_",	"_SQRT_"},
   {"TAN",	"_HTAN_",	"_TAN_"},
   {"TANH",	"_HTANH_",	"_TANH_"}

# endif

};
